#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from libadcc import HartreeFockProvider

from .DataHfProvider import DataOperatorIntegralProvider


class HfCounterData(HartreeFockProvider):
    """
    This class provides valid dummy data for any set of parameters
    given upon construction. This data can be imported and verified
    easily afterwards. All data is based on pain indices.
    """
    def __init__(self, n_alpha, n_beta, n_bas, n_orbs_alpha, restricted):
        if n_alpha != n_beta and restricted:
            raise ValueError("ROHF should not be tested")

        self.__n_alpha = n_alpha
        self.__n_beta = n_beta
        self.__n_bas = n_bas
        self.__n_orbs_alpha = n_orbs_alpha
        self.__restricted = restricted
        self.__mul = 10
        self.operator_integral_provider = DataOperatorIntegralProvider()

        # Do not forget the next line,
        # otherwise weird errors result
        super().__init__()

    def mul(self, n):
        return np.power(self.__mul, n)

    def get_backend(self):
        return "counterdata"

    def get_conv_tol(self):
        return 1e-10

    def get_restricted(self):
        return self.__restricted

    def get_energy_scf(self):
        return -1.0

    def get_spin_multiplicity(self):
        return 1

    def get_n_orbs_alpha(self):
        return self.__n_orbs_alpha

    def get_n_bas(self):
        return self.__n_bas

    def get_fa_range(self):
        return np.arange(1, self.__n_orbs_alpha + 1)

    def get_fb_range(self):
        if self.__restricted:
            return self.get_fa_range()
        else:
            return self.__n_orbs_alpha + self.get_fa_range()

    def get_f_range(self):
        return (self.get_fa_range(), self.get_fb_range())

    def get_b_range(self):
        return np.arange(1, self.__n_bas + 1)

    def fold_fock(self, range1, range2):
        """
        Build a reference fock matrix from two reference axes, where
        in each axis the alpha and the beta values are given as a tuple.
        """
        na1 = len(range1[0])  # Number of alphas for first axis
        na2 = len(range2[0])  # Number of alphas for second axis
        n1 = len(range1[0]) + len(range1[1])  # Total number for first axis
        n2 = len(range2[0]) + len(range2[1])  # Total number for first axis
        res = np.zeros((n1, n2))
        for (i, vi) in enumerate(np.hstack(range1)):
            for (j, vj) in enumerate(np.hstack(range2)):
                if i < na1 and j >= na2:
                    continue  # ab block is zero
                if i >= na1 and j < na2:
                    continue  # ba block is zero
                if vi < vj:
                    res[i, j] = self.mul(1) * vi + vj
                else:
                    res[i, j] = vi + self.mul(1) * vj
        return res

    def fill_orbcoeff_fb(self, out):
        out[:] = (np.hstack(self.get_f_range())[:, None] * self.mul(1)
                  + self.get_b_range()[None, :])

    def fill_occupation_f(self, out):
        n_oa = self.__n_orbs_alpha
        out[:] = np.zeros(2 * n_oa)
        out[:self.__n_alpha] = 1.
        out[n_oa:n_oa + self.__n_beta] = 1.

    def fill_orben_f(self, out):
        out[:] = np.hstack(self.get_f_range())

    def fill_fock_ff(self, slices, out):
        out[:] = self.fold_fock(self.get_f_range(), self.get_f_range())[slices]

    def fill_eri_ffff(self, slices, out):
        raise NotImplementedError("eri_ffff")

    def has_eri_phys_asym_ffff(self):
        return True

    def fold_eri(self, range1, range2, range3, range4):
        # Notice: This function does not yet work fully
        #         and is super slow, too.
        # The idea is to generate a nice ERI tensor. The symmetry-checked
        # import does not yet complain, but the resulting data does not agree

        na1 = len(range1[0])  # Number of alphas for first axis
        na2 = len(range2[0])  # Number of alphas for second axis
        na3 = len(range3[0])
        na4 = len(range4[0])
        n_alphas = [na1, na2, na3, na4]

        # Full ranges per axis
        rfulls = (np.hstack(range1), np.hstack(range2),
                  np.hstack(range3), np.hstack(range4))

        # Total number of orbitals per axis
        n_orbs = tuple(len(rfulls[i]) for i in range(4))

        # Function to compute one eri value. Takes a block and 4 indices
        # and reads rfulls and n_alphas
        def compute_eri_value(block, i1, i2, i3, i4):
            idx = np.array([i1, i2, i3, i4])
            val = np.array([rfulls[i][idx[i]] for i in range(4)])

            # map equivalent spin blocks onto another
            fac = 1.0
            if self.__restricted:
                if block == "bbbb":    # bbbb -> aaaa
                    block = "aaaa"
                elif block == "baba":  # baba -> abab
                    block = "abab"
                elif block == "abba":  # abba -> -baba -> -abab
                    block = "abab"
                    fac *= -1
                elif block == "baab":  # baab -> abba -> -baba -> -abab
                    block = "abab"
                    fac *= -1
                assert block in ("aaaa", "abab")

            # Zero elements by index permutation
            if val[0] == val[1]:
                return 0
            if val[2] == val[3]:
                return 0

            # Canonicalise the values wrt. index permutations
            if val[0] > val[1]:  # deal with ijkl = -jikl
                val[1], val[0] = val[0], val[1]
                fac *= -1
            if val[2] > val[3]:  # deal with ijkl = -ijlk
                val[2], val[3] = val[3], val[2]
                fac *= -1
            if val[0] > val[2]:  # deal with ijkl = klij
                val[0], val[1], val[2], val[3] = \
                    val[2], val[3], val[0], val[1]

            return fac * (
                + self.mul(3) * val[0] + self.mul(2) * val[1]
                + self.mul(1) * val[2] + self.mul(0) * val[3]
            )

        # Run only over blocks, which are non-zero by spin
        res = np.zeros(n_orbs)
        for block in ["aaaa", "abab", "abba", "bbbb", "baba", "baab"]:
            # Build the ranges and iterate over them:
            ranges = []
            for i, s in enumerate(block):
                if s == "a":
                    ranges.append(range(n_alphas[i]))
                else:
                    ranges.append(range(n_alphas[i], n_orbs[i]))

            for i1 in ranges[0]:
                for i2 in ranges[1]:
                    for i3 in ranges[2]:
                        for i4 in ranges[3]:
                            res[i1, i2, i3, i4] = compute_eri_value(
                                block, i1, i2, i3, i4
                            )
        return res

    def fill_eri_phys_asym_ffff(self, slices, out):
        full = self.fold_eri(self.get_f_range(), self.get_f_range(),
                             self.get_f_range(), self.get_f_range())
        out[:] = full[slices]



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from adcc import evaluate, lincomb


class GramSchmidtOrthogonaliser:
    def __init__(self, explicit_symmetrisation=None, n_rounds=1):
        """
        Initialise the GramSchmidtOrthogonaliser

        Parameters
        ----------
        explicit_symmetrisation
            The explicit symmetrisation to use on each orthogonalised vector
        n_rounds : int
            The number of times to apply the (regular) Gram-Schmidt
        """
        self.explicit_symmetrisation = explicit_symmetrisation
        self.n_rounds = n_rounds

    def qr(self, vectors):
        """
        A simple (and inefficient / inaccurate) QR decomposition based
        on Gram-Schmidt. Use only if no alternatives.

        vectors : list
            List of vectors representing the input matrix to decompose.
        """
        if len(vectors) == 0:
            return []
        elif len(vectors) == 1:
            norm_v = np.sqrt(vectors[0] @ vectors[0])
            return [evaluate(vectors[0] / norm_v)], np.array([[norm_v]])
        else:
            n_vec = len(vectors)
            Q = self.orthogonalise(vectors)
            R = np.zeros((n_vec, n_vec))
            for i in range(n_vec):
                for j in range(i, n_vec):
                    R[i, j] = Q[i] @ vectors[j]
            return Q, R

    def orthogonalise(self, vectors):
        """
        Orthogonalise the passed vectors with each other and return
        orthonormal vectors.
        """
        if len(vectors) == 0:
            return []
        subspace = [evaluate(vectors[0] / np.sqrt(vectors[0] @ vectors[0]))]
        for v in vectors[1:]:
            w = self.orthogonalise_against(v, subspace)
            subspace.append(evaluate(w / np.sqrt(w @ w)))
        return subspace

    def orthogonalise_against(self, vector, subspace):
        """
        Orthogonalise the passed vector against a subspace. The latter is assumed
        to only consist of orthonormal vectors. Effectively computes
        ``(1 - SS * SS^T) * vector`.

        vector
            Vector to make orthogonal to the subspace
        subspace : list
            Subspace of orthonormal vectors.
        """
        # Project out the components of the current subspace
        # That is form (1 - SS * SS^T) * vector = vector + SS * (-SS^T * vector)
        for _ in range(self.n_rounds):
            coefficients = np.hstack(([1], -(vector @ subspace)))
            vector = lincomb(coefficients, [vector] + subspace, evaluate=True)
            if self.explicit_symmetrisation is not None:
                self.explicit_symmetrisation.symmetrise(vector)
        return vector



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc


class Symmetry(libadcc.Symmetry):
    def __init__(self, mospaces, space, permutations=None,
                 spin_block_maps=None, spin_blocks_forbidden=None):
        if not isinstance(mospaces, libadcc.MoSpaces):
            raise TypeError("mospaces needs to be an MoSpaces instance.")

        super().__init__(mospaces, space)
        if permutations is not None:
            self.permutations = permutations
        if spin_block_maps is not None:
            self.spin_block_maps = spin_block_maps
        if spin_blocks_forbidden is not None:
            self.spin_blocks_forbidden = spin_blocks_forbidden

    def _repr_pretty_(self, pp, cycle):
        pp.text(self.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import numpy as np

from data import coeff_data, dip_data, eri_data, orben_data


def import_data():
    n_orbs_alpha = n_bas = 7
    n_orbs = 2 * n_orbs_alpha
    data = {
        "n_orbs_alpha": n_orbs_alpha,
        "n_bas": n_bas,
        "energy_scf": -7.4959319286025718e+01,
        "restricted": True,
        "threshold": 1e-12,
        "spin_multiplicity": 1,
        "multipoles": {
            "elec_0": -10,
            "nuclear_0": 10,
            "nuclear_1": np.array([1.693194615993441, 0.,
                                   1.196196642772152]),
            "elec_1": np.array(dip_data).reshape(3, n_bas, n_bas)
        },
    }
    data["occupation_f"] = np.array(5 * [1] + [0, 0] + 5 * [1] + [0, 0.])
    data["orbcoeff_fb"] = np.array(coeff_data).reshape((n_orbs, n_bas))
    data["orben_f"] = np.array(orben_data).reshape((n_orbs))
    data["eri_ffff"] = np.array(eri_data).reshape((n_orbs, n_orbs,
                                                   n_orbs, n_orbs))

    data["fock_ff"] = np.zeros((n_orbs, n_orbs))
    for i in range(n_orbs):
        data["fock_ff"][i, i] = orben_data[i]
    return data



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import time
import subprocess

sys.path.insert(0, os.path.abspath('..'))

import adcc  # noqa: E402

# -- Project information -----------------------------------------------------


def determine_tag():
    commit = subprocess.check_output("git rev-parse HEAD".split(),
                                     universal_newlines=True).strip()
    tag = subprocess.check_output("git tag --points-at".split() + [f"{commit}"],
                                  universal_newlines=True).strip()
    return tag


project = 'adcc'
author = ", ".join(adcc.__authors__)
copyright = time.strftime("%Y") + ", " + author
version = adcc.__version__
release = adcc.__version__

if determine_tag() == "":
    version = "dev"
    release = adcc.__version__ + "+dev"

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.imgmath',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    # 'sphinx.ext.doctest',
    # 'sphinx.ext.todo',
    # 'sphinx.ext.coverage',
    #
    'sphinx_automodapi.automodapi',
    'sphinxcontrib.bibtex',
    'breathe'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# A nice logo. Should not exceed 200px
html_logo = None

# The canonical HTML url
html_baseurl = "https://adc-connect.org/"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "canonical_url": html_baseurl,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'adccdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'adcc.tex', 'adcc Documentation', author, 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'adcc', 'adcc Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'adcc', 'adcc Documentation',
     author, 'adcc', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# Imgmath settings
imgmath_latex_preamble = "".join([
    r"\newcommand*{\mat}[1]{\ensuremath{\mathbf{#1}}}",
    r"\renewcommand*{\vec}[1]{\ensuremath{\underline{\boldsymbol{#1}}}}"
    r"\DeclareMathOperator{\tr}{tr}",
    #
    r"\newcommand*{\Op}[1]{\ensuremath{\hat{\mathcal{#1}}}}",
    r"\newcommand*{\bra}[1]{\left\langle#1\right|}",
    r"\newcommand*{\ket}[1]{\left|#1\right\rangle}"
    r"\newcommand*{\braket}[2]{\left\langle#1\middle|#2\right\rangle}"
])

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True

# __init__ docstring appears in documentation
autoclass_content = "init"

# Breathe settings
libadcc_doc_dir = (
    os.path.abspath(os.path.dirname(__file__))
    + "/../build/libadcc_docs/xml"
)
breathe_projects = {"libadcc": libadcc_doc_dir, }
breathe_default_project = "libadcc"

# Bibtex
bibtex_bibfiles = ['ref.bib', 'pub.bib']



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from adcc import AmplitudeVector, Symmetry, Tensor

from ..AdcMatrix import AdcMatrixlike


def guess_zero(matrix, spin_change=0, spin_block_symmetrisation="none"):
    """
    Return an AmplitudeVector object filled with zeros, but where the symmetry
    has been properly set up to meet the specified requirements on the guesses.

    matrix       The matrix for which guesses are to be constructed
    spin_change  The spin change to enforce in an excitation.
                 Typical values are 0 (singlet/triplet/any) and -1 (spin-flip).
    spin_block_symmetrisation
                 Symmetrisation to enforce between equivalent spin blocks, which
                 all yield the desired spin_change. E.g. if spin_change == 0,
                 then both the alpha->alpha and beta->beta blocks of the singles
                 part of the excitation vector achieve a spin change of 0.
                 The symmetry specified with this parameter will then be imposed
                 between the a-a and b-b blocks. Valid values are "none",
                 "symmetric" and "antisymmetric", where "none" enforces
                 no particular symmetry.
    """
    return AmplitudeVector(**{
        block: Tensor(sym) for block, sym in guess_symmetries(
            matrix, spin_change=spin_change,
            spin_block_symmetrisation=spin_block_symmetrisation
        ).items()
    })


def guess_symmetries(matrix, spin_change=0, spin_block_symmetrisation="none"):
    """
    Return guess symmetry objects (one for each AmplitudeVector block) such
    that the specified requirements on the guesses are satisfied.

    matrix       The matrix for which guesses are to be constructed
    spin_change  The spin change to enforce in an excitation.
                 Typical values are 0 (singlet/triplet/any) and -1 (spin-flip).
    spin_block_symmetrisation
                 Symmetrisation to enforce between equivalent spin blocks, which
                 all yield the desired spin_change. E.g. if spin_change == 0,
                 then both the alpha->alpha and beta->beta blocks of the singles
                 part of the excitation vector achieve a spin change of 0.
                 The symmetry specified with this parameter will then be imposed
                 between the a-a and b-b blocks. Valid values are "none",
                 "symmetric" and "antisymmetric", where "none" enforces
                 no particular symmetry.
    """
    if not isinstance(matrix, AdcMatrixlike):
        raise TypeError("matrix needs to be of type AdcMatrixlike")
    if spin_block_symmetrisation not in ["none", "symmetric", "antisymmetric"]:
        raise ValueError("Invalid value for spin_block_symmetrisation: "
                         "{}".format(spin_block_symmetrisation))
    if spin_block_symmetrisation != "none" and \
       not matrix.reference_state.restricted:
        raise ValueError("spin_block_symmetrisation != none is only valid for "
                         "ADC calculations on top of restricted reference "
                         "states.")
    if int(spin_change * 2) / 2 != spin_change:
        raise ValueError("Only integer or half-integer spin_change is allowed. "
                         "You passed {}".format(spin_change))

    max_spin_change = 0
    if "ph" in matrix.axis_blocks:
        max_spin_change = 1
    if "pphh" in matrix.axis_blocks:
        max_spin_change = 2
    if spin_change > max_spin_change:
        raise ValueError("spin_change for singles guesses may only be in the "
                         f"range [{-max_spin_change}, {max_spin_change}] and "
                         f"not {spin_change}.")

    symmetries = {}
    if "ph" in matrix.axis_blocks:
        symmetries["ph"] = guess_symmetry_singles(
            matrix, spin_change=spin_change,
            spin_block_symmetrisation=spin_block_symmetrisation
        )
    if "pphh" in matrix.axis_blocks:
        symmetries["pphh"] = guess_symmetry_doubles(
            matrix, spin_change=spin_change,
            spin_block_symmetrisation=spin_block_symmetrisation
        )
    return symmetries


def guess_symmetry_singles(matrix, spin_change=0,
                           spin_block_symmetrisation="none"):
    symmetry = Symmetry(matrix.mospaces, "".join(matrix.axis_spaces["ph"]))
    symmetry.irreps_allowed = ["A"]
    if spin_change != 0 and spin_block_symmetrisation != "none":
        raise NotImplementedError("spin_symmetrisation != 'none' only "
                                  "implemented for spin_change == 0")
    elif spin_block_symmetrisation == "symmetric":
        symmetry.spin_block_maps = [("aa", "bb", 1)]
        symmetry.spin_blocks_forbidden = ["ab", "ba"]
    elif spin_block_symmetrisation == "antisymmetric":
        symmetry.spin_block_maps = [("aa", "bb", -1)]
        symmetry.spin_blocks_forbidden = ["ab", "ba"]
    return symmetry


def guess_symmetry_doubles(matrix, spin_change=0,
                           spin_block_symmetrisation="none"):
    spaces_d = matrix.axis_spaces["pphh"]
    symmetry = Symmetry(matrix.mospaces, "".join(spaces_d))
    symmetry.irreps_allowed = ["A"]

    if spin_change != 0 and spin_block_symmetrisation != "none":
        raise NotImplementedError("spin_symmetrisation != 'none' only "
                                  "implemented for spin_change == 0")

    if spin_change == 0 \
       and spin_block_symmetrisation in ("symmetric", "antisymmetric"):
        fac = 1 if spin_block_symmetrisation == "symmetric" else -1
        # Spin mapping between blocks where alpha and beta are just mirrored
        symmetry.spin_block_maps = [("aaaa", "bbbb", fac),
                                    ("abab", "baba", fac),
                                    ("abba", "baab", fac)]

        # Mark blocks which change spin as forbidden
        symmetry.spin_blocks_forbidden = ["aabb",  # spin_change +2
                                          "bbaa",  # spin_change -2
                                          "aaab",  # spin_change +1
                                          "aaba",  # spin_change +1
                                          "abaa",  # spin_change -1
                                          "baaa",  # spin_change -1
                                          "abbb",  # spin_change +1
                                          "babb",  # spin_change +1
                                          "bbab",  # spin_change -1
                                          "bbba"]  # spin_change -1

    # Add index permutation symmetry:
    permutations = ["ijab"]
    if spaces_d[0] == spaces_d[1]:
        permutations.append("-jiab")
    if spaces_d[2] == spaces_d[3]:
        permutations.append("-ijba")
    if len(permutations) > 1:
        symmetry.permutations = permutations
    return symmetry



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np
from scipy import constants

from .import adc_pp
from .AdcMethod import AdcType
from .ElectronicStates import TableColumn
from .ElectronicTransition import ElectronicTransition
from .functions import dot


class ExcitedStates(ElectronicTransition):
    _module = adc_pp

    def __init__(self, data, method: str = None, property_method: str = None):
        super().__init__(data, method, property_method)

        if self.method.adc_type is not AdcType.PP:
            raise ValueError("ExcitedStates computes excited state properties "
                             "for PP-ADC. Got the non-PP-ADC method "
                             f"{self.method.name}")

    def _repr_pretty_(self, pp, cycle):
        if cycle:
            pp.text("ExcitedStates(...)")
        else:
            pp.text(self.describe())

    def describe(self, oscillator_strengths=True, rotatory_strengths=False,
                 state_dipole_moments=False, transition_dipole_moments=False,
                 block_norms=True, ssq=False):
        """
        Return a string providing a human-readable description of the class

        Parameters
        ----------
        oscillator_strengths : bool optional
            Show oscillator strengths, by default ``True``.

        rotatory_strengths : bool optional
           Show rotatory strengths, by default ``False``.

        state_dipole_moments : bool, optional
            Show state dipole moments, by default ``False``.

        transition_dipole_moments : bool, optional
            Show state dipole moments, by default ``False``.

        block_norms : bool, optional
            Show the norms of the (1p1h, 2p2h, ...) blocks of the excited states,
            by default ``True``.

        ssq : bool, optional
            Show the <S^2> values of the excited states,  by default ``False``.
        """
        has_dipole = "electric_dipole" in self.operators.available
        has_rotatory = all(
            op in self.operators.available for op in
            ["magnetic_dipole", "electric_dipole_velocity"]
        )
        # Collect the columns to print
        columns: list[TableColumn] = []
        values: list[str] = []
        # count the number of states
        values.extend(str(i) for i in range(self.size))
        columns.append(TableColumn(header="#", values=values.copy(), unit=""))
        values.clear()
        # excitation energy in a.u. and eV
        eV = constants.value("Hartree energy in eV")
        values.extend(f"{e:^13.7g} {e * eV:^13.7g}" for e in self.excitation_energy)
        columns.append(TableColumn(
            header="excitation energy", values=values.copy(),
            unit="(au)         (eV)"
        ))
        values.clear()
        # the transition dipole moments
        if transition_dipole_moments and has_dipole:
            for tdm in self.transition_dipole_moment:
                tdmx, tdmy, tdmz = tdm
                values.append(
                    f"{tdmx:^8.4f} {tdmy:^8.4f} {tdmz:^8.4f}"
                    f"{np.linalg.norm(tdm):^8.4f}"
                )
            columns.append(TableColumn(
                header="transition dipole moment", values=values.copy(),
                unit="x(au)    y(au)    z(au)    abs(au)"
            ))
            values.clear()
        # the oscillator strengths
        if oscillator_strengths and has_dipole:
            values.extend(f"{osc:^8.4f}" for osc in self.oscillator_strength)
            columns.append(TableColumn(
                header="osc str", values=values.copy(), unit="(au)"
            ))
            values.clear()
        if rotatory_strengths and has_rotatory:
            values.extend(f"{rot:^8.4f}" for rot in self.rotatory_strength)
            columns.append(TableColumn(
                header="rot str", values=values.copy(), unit="(au)"
            ))
            values.clear()
        # vector norm
        if block_norms and "ph" in self.matrix.axis_blocks:
            values.extend(f"{dot(vec.ph, vec.ph):^9.4f}"
                          for vec in self.excitation_vector)
            columns.append(TableColumn(
                header="|v1|^2", values=values.copy(), unit=""
            ))
            values.clear()
        if block_norms and "pphh" in self.matrix.axis_blocks:
            values.extend(f"{dot(vec.pphh, vec.pphh):^9.4f}"
                          for vec in self.excitation_vector)
            columns.append(TableColumn(
                header="|v2|^2", values=values.copy(), unit=""
            ))
            values.clear()
        # the state dipole moment
        if state_dipole_moments and has_dipole:
            for dm in self.state_dipole_moment:
                dmx, dmy, dmz = dm
                values.append(
                    f"{dmx:^8.4f} {dmy:^8.4f} {dmz:^8.4f}"
                    f"{np.linalg.norm(dm):^8.4f}"
                )
            columns.append(TableColumn(
                header="state dipole moment", values=values.copy(),
                unit="x(au)    y(au)    z(au)    abs(au)"
            ))
            values.clear()
            values.clear()
        if ssq and not self.reference_state.restricted:
            values.extend(f"{ssq:^9.4f}"
                          for ssq in self.state_ssq)
            columns.append(TableColumn(
                header="<S^2>", values=values.copy(), unit="(au)"
            ))
            values.clear()
        return self._describe(columns)

    def to_qcvars(self, properties=False, recurse=False):
        """
        Return a dictionary with property keys compatible to a Psi4 wavefunction
        or a QCEngine Atomicresults object.
        """
        name = self.method.name.upper()

        qcvars = {
            "EXCITATION KIND": self.kind.upper(),
            "NUMBER OF EXCITED STATES": len(self.excitation_energy),
            f"{name} ITERATIONS": self.n_iter,
            f"{name} EXCITATION ENERGIES": self.excitation_energy,
        }

        if properties:
            qcvars.update({
                # Transition properties
                f"{name} TRANSITION DIPOLES (LEN)": self.transition_dipole_moment,
                f"{name} TRANSITION DIPOLES (VEL)": self.transition_dipole_moment_velocity,  # noqa: E501
                f"{name} OSCILLATOR STRENGTHS (LEN)": self.oscillator_strength,
                f"{name} OSCILLATOR STRENGTHS (VEL)": self.oscillator_strength_velocity,  # noqa: E501
                f"{name} ROTATIONAL STRENGTHS (VEL)": self.rotatory_strength,
                #
                # State properties
                f"{name} STATE DIPOLES": self.state_dipole_moment
            })

        if recurse:
            mpvars = self.ground_state.to_qcvars(properties, recurse=True,
                                                 maxlevel=self.method.level)
            qcvars.update(mpvars)
        return qcvars



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import psi4

# Run SCF in psi4
mol = psi4.geometry("""
    0 3
    H 0 0 0
    F 0 0 2.5
    symmetry c1
    units au
    no_reorient
    no_com
    """)

psi4.set_num_threads(adcc.get_n_threads())
psi4.core.be_quiet()
psi4.set_options({'basis': "6-31g",
                  'e_convergence': 1e-14,
                  'd_convergence': 1e-9,
                  'reference': 'uhf'})
scf_e, wfn = psi4.energy('SCF', return_wfn=True)

# Run solver and print results
states = adcc.adc2(wfn, n_spin_flip=5, conv_tol=1e-8)
print(states.describe())
print(states.describe_amplitudes())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------


class AmplitudeVector(dict):
    def __init__(self, **kwargs):
        """
        Construct an AmplitudeVector. Typical use cases are
        ``AmplitudeVector(ph=tensor_singles, pphh=tensor_doubles)``.
        """
        super().__init__(**kwargs)

    def __getattr__(self, key):
        if self.__contains__(key):
            return self.__getitem__(key)
        raise AttributeError

    def __setattr__(self, key, item):
        if self.__contains__(key):
            return self.__setitem__(key, item)
        raise AttributeError

    @property
    def blocks(self):
        """
        Return the blocks which are used inside the vector.
        """
        return sorted(self.keys())

    def copy(self):
        """Return a copy of the AmplitudeVector"""
        return AmplitudeVector(**{k: t.copy() for k, t in self.items()})

    def evaluate(self):
        for t in self.values():
            t.evaluate()
        return self

    @property
    def needs_evaluation(self):
        return any(t.needs_evaluation for _, t in self.items())

    def ones_like(self):
        """Return an empty AmplitudeVector of the same shape and symmetry"""
        return AmplitudeVector(**{k: t.ones_like() for k, t in self.items()})

    def empty_like(self):
        """Return an empty AmplitudeVector of the same shape and symmetry"""
        return AmplitudeVector(**{k: t.empty_like() for k, t in self.items()})

    def nosym_like(self):
        """Return an empty AmplitudeVector of the same shape and symmetry"""
        return AmplitudeVector(**{k: t.nosym_like() for k, t in self.items()})

    def zeros_like(self):
        """Return an AmplitudeVector of the same shape and symmetry with
           all elements set to zero"""
        return AmplitudeVector(**{k: t.zeros_like() for k, t in self.items()})

    def set_random(self):
        for t in self.values():
            t.set_random()
        return self

    def dot(self, other):
        """Return the dot product with another AmplitudeVector
        or the dot products with a list of AmplitudeVectors.
        In the latter case a np.ndarray is returned.
        """
        if isinstance(other, list):
            # Make a list where the first index is all singles parts,
            # the second is all doubles parts and so on
            return sum(self[b].dot([av[b] for av in other]) for b in self.keys())
        else:
            return sum(self[b].dot(other[b]) for b in self.keys())

    def __matmul__(self, other):
        if isinstance(other, AmplitudeVector):
            return self.dot(other)
        if isinstance(other, list):
            if all(isinstance(elem, AmplitudeVector) for elem in other):
                return self.dot(other)
        return NotImplemented

    def __forward_to_blocks(self, fname, other):
        if isinstance(other, AmplitudeVector):
            if sorted(other.blocks) != sorted(self.blocks):
                raise ValueError("Blocks of both AmplitudeVector objects "
                                 f"need to agree to perform {fname}")
            ret = {k: getattr(tensor, fname)(other[k])
                   for k, tensor in self.items()}
        else:
            ret = {k: getattr(tensor, fname)(other) for k, tensor in self.items()}
        if any(r == NotImplemented for r in ret.values()):
            return NotImplemented
        else:
            return AmplitudeVector(**ret)

    def __mul__(self, other):
        return self.__forward_to_blocks("__mul__", other)

    def __rmul__(self, other):
        return self.__forward_to_blocks("__rmul__", other)

    def __sub__(self, other):
        return self.__forward_to_blocks("__sub__", other)

    def __rsub__(self, other):
        return self.__forward_to_blocks("__rsub__", other)

    def __truediv__(self, other):
        return self.__forward_to_blocks("__truediv__", other)

    def __imul__(self, other):
        return self.__forward_to_blocks("__imul__", other)

    def __iadd__(self, other):
        return self.__forward_to_blocks("__iadd__", other)

    def __isub__(self, other):
        return self.__forward_to_blocks("__isub__", other)

    def __itruediv__(self, other):
        return self.__forward_to_blocks("__itruediv__", other)

    def __repr__(self):
        return "AmplitudeVector(" + "=..., ".join(self.blocks) + "=...)"

    # __add__ is special because we want to be able to add AmplitudeVectors
    # with missing blocks
    def __add__(self, other):
        if isinstance(other, AmplitudeVector):
            allblocks = sorted(set(self.blocks).union(other.blocks))
            ret = {k: self.get(k, 0) + other.get(k, 0) for k in allblocks}
            ret = {k: v for k, v in ret.items() if v != 0}
        else:
            ret = {k: tensor + other for k, tensor in self.items()}
        return AmplitudeVector(**ret)

    def __radd__(self, other):
        if isinstance(other, AmplitudeVector):
            return other.__add__(self)
        else:
            ret = {k: other + tensor for k, tensor in self.items()}
            return AmplitudeVector(**ret)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2025-2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from typing import (
    Callable, Optional, TextIO, Protocol, Deque, TypeVar, Generic, cast
)
from collections import deque
import sys

import numpy as np


class DIISError(Exception):
    """is the base class of errors visible from this module."""
    pass


class SubspaceError(DIISError):
    """is raised when encountering problems related to the subspace setup."""
    pass


Vector = TypeVar("Vector", bound="DIISVector")


class DIISVector(Protocol):
    def dot(self: Vector, other: Vector) -> float:
        ...

    def zeros_like(self: Vector) -> Vector:
        ...

    def __sub__(self: Vector, other: Vector) -> Vector:
        ...

    def __mul__(self: Vector, other: float) -> Vector:
        ...

    def __iadd__(self: Vector, other: Vector) -> Vector:
        ...


class DIISCallback(Protocol):
    def __call__(self, state: "DIISSubspace", identifier: str,
                 file: TextIO = sys.stdout) -> None:
        ...


class DIISSubspace(Generic[Vector]):
    """
    Manages the DIIS matrix and the subspace and error vectors.

    Parameters
    ----------
    max_size: int
        The maximum number of vectors to keep in the subspace simultaneously.
    start_size: int
        The minimal number of subspace vectors required in the subspace
        in order to start extrapolating. If less vectors are in the
        subspace, linear steps will be performed.
    """
    def __init__(self, max_size: int, start_size: int):
        if max_size < 2:
            raise SubspaceError(f"The DIIS maximum subspace size ({max_size=}) "
                                "has to be greater than 1.")
        if start_size < 2:
            raise SubspaceError(f"The DIIS start size ({start_size=}) has to "
                                "be greater than 1.")
        if start_size > max_size:
            raise SubspaceError(f"The maximum subspace size ({max_size=}) has to "
                                "be larger than the DIIS start size "
                                f"({start_size=}).")

        self.max_size: int = max_size
        self.start_size: int = start_size
        self.subspace_vectors: Deque[Vector] = deque(maxlen=max_size)
        self.error_vectors: Deque[Vector] = deque(maxlen=max_size)
        self.overlap: np.ndarray[tuple[int, int], np.dtype[np.float64]] = np.zeros(
            (0, 0), dtype=np.float64
        )
        self.step_info: Optional[str] = None
        self.converged: bool = False
        self.n_iter: int = 0

    def is_converged(self, conv_tol: float) -> bool:
        """
        Computes the Frobenius norm of the last error vector and checks
        whether it is smaller than the convergence criterion passed as
        conv_tol.
        """
        if self.n_iter >= 1 and self.residual_norm < conv_tol:
            self.converged = True
        return self.converged

    @property
    def residual_norm(self) -> np.float64:
        """
        Returns the Frobenius norm of the most recent error vector, which is
        just the square root of the last diagonal element of the current
        overlap matrix.
        """
        return np.sqrt(self.overlap[-1, -1])

    @property
    def size(self) -> int:
        """Returns the current subspace size"""
        return len(self.subspace_vectors)

    @property
    def last_vector(self) -> Vector:
        """
        Returns the most recent subspace vector, that is, in case of
        convergence, the solution vector.
        """
        return self.subspace_vectors[-1]

    def add(self, subspace_vector: Vector, error_vector: Vector) -> None:
        """
        Adds a new vector and corresponding error vector to the subspace and
        updates the DIIS matrix. Also the iteration counter is incremented.
        """
        self.n_iter += 1
        self.subspace_vectors.append(subspace_vector)
        self.error_vectors.append(error_vector)
        self._update_overlap()

    def _update_overlap(self) -> None:
        """
        Computes the matrix of error vectors in the current subspace.
        Only new elements are computed, elements which have been previously
        computed are transferred from the old overlap matrix.

        Here, two cases can occur. The dimension of the currently stored overlap
        matrix of the previous step ...
        1) equals the subspace size of the current step, that is, the subspace has
           already reached its maximum size in the previous step. In this case,
           the oldest entries (first row and column) are discarded.
        2) is one smaller than the subspace size of the current step, that is, the
           subspace has grown by one. In this case, the overlap matrix is extended
           to the right and bottom.

        The logic of this function is written in a general way in order to
        simultaneously cover both cases.
        """
        if not self.error_vectors:
            raise SubspaceError("No error vectors in subspace.")

        cur_size = len(self.error_vectors)
        copy_size = cur_size - 1
        prev_size = self.overlap.shape[0]
        assert prev_size == copy_size or prev_size == cur_size
        # init new overlap matrix, copy elements and add new elements
        new_overlap = np.zeros((cur_size, cur_size), dtype=np.float64)
        new_overlap.fill(float("NaN"))
        if copy_size > 0:
            new_overlap[:copy_size, :copy_size] = (
                self.overlap[-copy_size:, -copy_size:]
            )
        for ind, error_vec in enumerate(self.error_vectors):
            overlap_value = error_vec.dot(self.error_vectors[-1])
            new_overlap[ind, -1] = overlap_value
            # create transposed element if this is not the diagonal element
            if ind != cur_size - 1:
                new_overlap[-1, ind] = overlap_value
        # Check that there are no NaN elements in the updated overlap matrix,
        # i.e., that all elements have been properly populated
        assert not np.any(np.isnan(new_overlap))
        self.overlap = new_overlap

    def compute_guess(self, n_omit_vectors: int = 0) -> Vector:
        """
        Computes a DIIS guess from the current subspace of size n by solving a
        system of linear equations A * x = b, where A is a matrix of size
        (n+1 x n+1) and b is a vector of size (n+1). Using the error overlap
        matrix elements Sij, the A and b quantities are set up according to

            (S00/S00 S01/S00 ... S0n/S00  -1  )           (  0  )
            (S10/S00 S11/S00 ... S1n/S00  -1  )           (  0  )
        A = (  ...     ...   ...   ...    ... )  and  b = ( ... )
            (Sn0/S00 Sn1/S00 ... Snn/S00  -1  )           (  0  )
            (   -1      -1   ...    -1     0  )           ( -1  )

        The guess is then computed as linear combination of the current subspace
        vectors, using the first (n - n_omit_vectors) elements of the solution
        vector x as coefficients.

        Returns the guess vector.

        Parameters
        ----------
        n_omit_vectors: int
            In the DIIS extrapolation, omit the n subspace vectors associated
            with the error vectors with the largest Frobenius norm.
        """
        if not self.size:
            raise SubspaceError("No vectors in subspace.")
        elif n_omit_vectors >= self.size:
            raise SubspaceError(f"Can not omit {n_omit_vectors} vectors "
                                f"because there are only {self.size} vectors "
                                "in the subspace.")

        # we only have a single vector -> perform a linear step
        if self.size == 1 or self.size - n_omit_vectors <= 1 or \
                self.size < self.start_size:
            self.step_info = "Linear step"
            return self.last_vector

        fill_value = -1.0
        diis_size = self.size
        # build the DIIS matrix
        A: np.ndarray[tuple[int, int], np.dtype[np.float64]] = np.zeros(
            (diis_size + 1, diis_size + 1), dtype=np.float64
        )
        A.fill(fill_value)
        A[-1, -1] = 0.0
        A[:diis_size, :diis_size] = self.overlap / self.overlap[0, 0]
        # build the right hand side
        b: np.ndarray[tuple[int], np.dtype[np.float64]] = np.zeros(
            diis_size + 1, dtype=np.float64
        )
        b[-1] = fill_value
        # determine the indices of vectors to discard
        inds_to_discard = []
        if n_omit_vectors > 0:
            inds_to_discard = np.argsort(self.overlap.diagonal())[-n_omit_vectors:]
            # shrink the system of linear equations by removing rows/columns
            # collected in inds_to_discard
            A = np.delete(A, inds_to_discard, axis=0)
            A = np.delete(A, inds_to_discard, axis=1)
            b = np.delete(b, inds_to_discard, axis=0)
        # solve the system of linear equations
        coefficients = np.linalg.solve(A, b)
        # discard the Lagrange multiplier, which is not needed for DIIS
        # extrapolation
        # also: the coefficients have to be a 1D array with dtype float64
        # the type checker unfortunately can not infer this automatically
        coefficients = cast(
            np.ndarray[tuple[int], np.dtype[np.float64]], coefficients[:-1]
        )
        # build the linear combination:
        # determine the indices of vectors to use for the generation of the
        # DIIS-extrapolated guess, i.e., vectors which have not beeen
        # discarded above
        vec_indices = [i for i in range(self.size) if i not in inds_to_discard]
        guess = self.subspace_vectors[0].zeros_like()
        assert len(vec_indices) == len(coefficients)
        for vec_ind, coeff in zip(vec_indices, coefficients):
            guess += self.subspace_vectors[vec_ind] * coeff

        self.step_info = (
            f"DIIS step from {len(vec_indices)} error vectors, subspace size: "
            f"{self.size}"
        )
        return guess


def default_print(state: DIISSubspace, identifier: str, file: TextIO = sys.stdout):
    if identifier == "start":
        print("Niter   DIIS_error  comment", file=file)
    elif identifier == "next_iter":
        fmt = "{n_iter:4d} {residual: >13.5E}  {step_info:s}"
        print(fmt.format(n_iter=state.n_iter,
                         residual=state.residual_norm,
                         step_info=state.step_info), file=file)
    elif identifier == "is_converged":
        print("=== Converged ===", file=file)


def _no_print(state: DIISSubspace, identifier: str, file: TextIO = sys.stdout):
    pass


def diis(updater: Callable[[Vector], Vector], guess_vector: Vector,
         diis_start_size: int = 3, max_subspace_size: int = 7,
         conv_tol: float = 1e-9, n_max_iterations: int = 100,
         callback: Optional[DIISCallback] = None) -> Vector:
    """
    Implementation of the direct inversion of the iterative subspace algorithm.

    Parameters
    ----------
    updater: callable
        Callable that takes the guess vector and updates it once.
    guess_vector
        The guess vector to start with.
    diis_start_size: int, optional
        Perform n linear steps until the DIIS subspace contains at least
        n vectors. Has to be smaller than or equal to max_subspace_size
        (default: 3).
    max_subspace_size: int, optional
        The maximum number of vectors to keep in the subspace simultaneously
        (default: 7).
    conv_tol: float, optional
        The convergence tolerance on the Frobenius norm of the
        error vector (default: 1e-9).
    n_max_iterations: int, optional
        The maximum number of allowed iterations (default: 100).
    callback: DIISCallback, optional
        A callable that is called after each iteration, e.g., to produce
        printout.
    """
    if callback is None:
        callback = _no_print
    if n_max_iterations < 1:
        raise DIISError(f"The maximum number of iterations ({n_max_iterations=}) "
                        "can not be smaller than 1.")

    # initialize DIIS subspace
    diis_subspace: DIISSubspace[Vector] = DIISSubspace(
        max_size=max_subspace_size, start_size=diis_start_size
    )
    # perform an initial linear step
    new_subspace_vector = updater(guess_vector)
    new_error_vector = new_subspace_vector - guess_vector
    diis_subspace.add(new_subspace_vector, new_error_vector)
    diis_subspace.step_info = "Linear step from guess"
    callback(diis_subspace, "start")
    callback(diis_subspace, "next_iter")
    # check for convergence (very unlikely)
    if diis_subspace.is_converged(conv_tol):
        callback(diis_subspace, "is_converged")
        return diis_subspace.last_vector

    # iterate
    while diis_subspace.n_iter < n_max_iterations:
        diis_guess = diis_subspace.compute_guess()
        new_subspace_vector = updater(diis_guess)
        new_error_vector = new_subspace_vector - diis_guess
        diis_subspace.add(new_subspace_vector, new_error_vector)
        callback(diis_subspace, "next_iter")

        # check for convergence
        if diis_subspace.is_converged(conv_tol):
            callback(diis_subspace, "is_converged")
            break
    else:  # no convergence achieved within n_max_iterations iterations
        raise DIISError(
            f"No convergence detected after {n_max_iterations} iterations"
        )
    return diis_subspace.last_vector



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .guess_zero import guess_symmetries, guess_zero
from .guesses_from_diagonal import guesses_from_diagonal

__all__ = ["guess_zero", "guesses_from_diagonal",
           "guesses_singlet", "guesses_triplet", "guesses_any",
           "guesses_spin_flip", "guess_symmetries"]


def guess_kwargs_kind(kind):
    """
    Return the kwargs required to be passed to `guesses_from_diagonal` to
    computed states of the passed excitation `kind`.
    """
    kwargsmap = dict(
        singlet=dict(spin_block_symmetrisation="symmetric", spin_change=0),
        triplet=dict(spin_block_symmetrisation="antisymmetric", spin_change=0),
        spin_flip=dict(spin_block_symmetrisation="none", spin_change=-1),
        any=dict(spin_block_symmetrisation="none", spin_change=0),
    )
    try:
        return kwargsmap[kind]
    except KeyError:
        raise ValueError(f"Kind not known: {kind}")


def guesses_singlet(matrix, n_guesses, block="ph", **kwargs):
    """
    Obtain guesses for computing singlet states by inspecting the passed
    ADC matrix.

    matrix       The matrix for which guesses are to be constructed
    n_guesses    The number of guesses to be searched for. Less number of
                 vectors are returned if this many could not be found.
    block        Diagonal block to use for obtaining the guesses
                 (typically "ph" or "pphh").
    kwargs       Any other argument understood by guesses_from_diagonal.
    """
    return guesses_from_diagonal(matrix, n_guesses, block=block,
                                 **guess_kwargs_kind("singlet"), **kwargs)


def guesses_triplet(matrix, n_guesses, block="ph", **kwargs):
    """
    Obtain guesses for computing triplet states by inspecting the passed
    ADC matrix.

    matrix       The matrix for which guesses are to be constructed
    n_guesses    The number of guesses to be searched for. Less number of
                 vectors are returned if this many could not be found.
    block        Diagonal block to use for obtaining the guesses
                 (typically "ph" or "pphh").
    kwargs       Any other argument understood by guesses_from_diagonal.
    """
    return guesses_from_diagonal(matrix, n_guesses, block=block,
                                 **guess_kwargs_kind("triplet"), **kwargs)


# guesses for computing any state (singlet or triplet)
guesses_any = guesses_from_diagonal


def guesses_spin_flip(matrix, n_guesses, block="ph", **kwargs):
    """
    Obtain guesses for computing spin-flip states by inspecting the passed
    ADC matrix.

    matrix       The matrix for which guesses are to be constructed
    n_guesses    The number of guesses to be searched for. Less number of
                 vectors are returned if this many could not be found.
    block        Diagonal block to use for obtaining the guesses
                 (typically "ph" or "pphh").
    kwargs       Any other argument understood by guesses_from_diagonal.
    """
    return guesses_from_diagonal(matrix, n_guesses, block=block,
                                 **guess_kwargs_kind("spin_flip"), **kwargs)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

import h5py

from libadcc import HartreeFockProvider


def get_scalar_value(data, key, default=None):
    """
    Function to allow to retrieve array data both from dict and from
    HDF5 objects
    """
    if "/" in key:
        key, subkey = key.split("/", 1)
        return get_scalar_value(data.get(key, {}), subkey, default=default)
    if default is not None and key not in data:
        return default

    value = data[key]
    if not hasattr(value, "shape"):
        return value  # Just a scalar
    elif value.shape == ():
        return value[()]
    elif value.shape == (1, ):
        return value[0]
    else:
        raise ValueError("Unrecognised scalar value shape ", value.shape,
                         " should be () or (1, )")


def get_array_value(data, key, default=None):
    """
    Function to allow to retrieve scalar data both from dict and from
    HDF5 objects
    """
    if "/" in key:
        key, subkey = key.split("/", 1)
        return get_array_value(data.get(key, {}), subkey, default=default)
    if default is not None and key not in data:
        return default
    return np.asarray(data[key])


class DataOperatorIntegralProvider:
    def __init__(self, backend="data"):
        self.backend = backend

    @property
    def available(self) -> tuple[str, ...]:
        blacklist = ("backend", "available")
        return tuple(
            integral for integral in dir(self)
            if not integral.startswith("_") and integral not in blacklist
        )


class DataHfProvider(HartreeFockProvider):
    def __init__(self, data):
        """
        Initialise the DataHfProvider class with the `data` being a supported
        data container (currently python dictionary or HDF5 file).
        Let `nf` denote the number of Fock spin orbitals (i.e. the sum of both
        the alpha and the beta orbitals) and `nb` the number of basis functions.
        With `array` we indicate either a `np.array` or an HDF5 dataset.
        The following keys are required in the container:

        1. **restricted** (`bool`): `True` for a restricted SCF calculation,
           `False` otherwise
        2. **conv_tol** (`float`): Tolerance value used for SCF convergence,
           should be roughly equivalent to l2 norm of the Pulay error.
        3. **orbcoeff_fb** (`.array` with dtype `float`, size `(nf, nb)`):
           SCF orbital coefficients, i.e. the uniform transform from the basis
           to the molecular orbitals.
        4. **occupation_f** (`array` with dtype `float`, size `(nf, )`:
           Occupation number for each SCF orbitals (i.e. diagonal of the HF
           density matrix in the SCF orbital basis).
        5. **orben_f** (`array` with dtype `float`, size `(nf, )`:
           SCF orbital energies
        6. **fock_ff** (`array` with dtype `float`, size `(nf, nf)`:
           Fock matrix in SCF orbital basis. Notice, the full matrix is expected
           also for restricted calculations.
        7. **eri_phys_asym_ffff** (`array` with dtype `float`,
           size `(nf, nf, nf, nf)`: Antisymmetrised electron-repulsion integral
           tensor in the SCF orbital basis, using the Physicists' indexing
           convention, i.e. that the index tuple `(i,j,k,l)` refers to
           the integral :math:`\\langle ij || kl \\rangle`, i.e.

           .. math::
              \\int_\\Omega \\int_\\Omega d r_1 d r_2 \\frac{
              \\phi_i(r_1) \\phi_j(r_2)
              \\phi_k(r_1) \\phi_l(r_2)}{|r_1 - r_2|}
              - \\int_\\Omega \\int_\\Omega d r_1 d r_2 \\frac{
              \\phi_i(r_1) \\phi_j(r_2)
              \\phi_l(r_1) \\phi_k(r_2)}{|r_1 - r_2|}

           The full tensor (including zero blocks) is expected.

        As an alternative to `eri_phys_asym_ffff`, the user may provide

        8. **eri_ffff** (`array` with dtype `float`, size `(nf, nf, nf, nf)`:
           Electron-repulsion integral tensor in chemists' notation.
           The index tuple `(i,j,k,l)` thus refers to the integral
           :math:`(ij|kl)`, which is

           .. math::
              \\int_\\Omega \\int_\\Omega d r_1 d r_2
              \\frac{\\phi_i(r_1) \\phi_j(r_1)
              \\phi_k(r_2) \\phi_l(r_2)}{|r_1 - r_2|}

           Notice, that no antisymmetrisation has been applied in this tensor.

        The above keys define the least set of quantities to start a calculation
        in `adcc`. In order to have access to properties such as dipole moments
        or to get the correct state energies, further keys are highly
        recommended to be provided as well.

        9. **energy_scf** (`float`): Final total SCF energy of both electronic
           and nuclear energy terms. (default: `0.0`)
        10. **multipoles**: Container with electric and nuclear
            multipole moments. Can be another dictionary or simply an HDF5
            group.

              - **elec_1** (`array`, size `(3, nb, nb)`):
                Electric dipole moment integrals in the atomic orbital basis
                (i.e. the discretisation basis with `nb` elements). First axis
                indicates cartesian component (x, y, z).
              - **nuc_0** (`float`): Total nuclear charge
              - **nuc_1** (`array` size `(3, )`: Nuclear dipole moment

            The defaults for all entries are all-zero multipoles.
        11. **spin_multiplicity** (`int`): The spin mulitplicity of the HF
            ground state described by the data. A value of `0` (for unknown)
            should be supplied for unrestricted calculations.
            (default: 1 for restricted and 0 for unrestricted calculations)

        A descriptive string for the backend can be supplied optionally as well.
        In case of using a python `dict` as the data container, this should be
        done using the key `backend`. For an HDF5 file, this should be done
        using the attribute `backend`. Defaults based on the filename are
        generated.

        Parameters
        ----------
        data : dict or h5py.File
            Dictionary containing the HartreeFock data to use. For the required
            keys see details above.
        """

        # Do not forget the next line, otherwise weird errors result
        super().__init__()
        self.data = data

        if isinstance(data, dict):
            self.__backend = data.get("backend", "dict")
        elif isinstance(data, h5py.File):
            if "r" not in data.mode:
                raise ValueError("Passed h5py.File stream (filename: {}) not "
                                 "readable.".format(data.filename))
            self.__backend = data.attrs.get(
                "backend", '<HDF5 file "{}">'.format(data.filename)
            )
        else:
            raise TypeError("Can only deal with data objects of type dict "
                            "or h5py.File.")

        if data["orbcoeff_fb"].shape[0] % 2 != 0:
            raise ValueError("orbcoeff_fb first axis should have even length")
        nb = self.get_n_bas()
        nf = 2 * self.get_n_orbs_alpha()

        checks = [("orbcoeff_fb", (nf, nb)), ("occupation_f", (nf, )),
                  ("orben_f", (nf, )), ("fock_ff", (nf, nf)),
                  ("eri_ffff", (nf, nf, nf, nf)),
                  ("eri_phys_asym_ffff", (nf, nf, nf, nf)), ]
        for key, exshape in checks:
            if key not in data:
                continue
            if data[key].shape != exshape:
                raise ValueError("Shape mismatch for key {}: Expected {}, but "
                                 "got {}.".format(key, exshape,
                                                  data[key].shape))

        # Setup integral data
        opprov = DataOperatorIntegralProvider(self.__backend)
        mmp = data.get("multipoles", {})
        if "elec_1" in mmp:
            if mmp["elec_1"].shape != (3, nb, nb):
                raise ValueError("multipoles/elec_1 is expected to have shape "
                                 + str((3, nb, nb)) + " not "
                                 + str(mmp["elec_1"].shape))
            opprov.electric_dipole = np.asarray(mmp["elec_1"])
        if "elec_2_origin" in mmp:
            def get_integral_elquad(gauge_origin):
                return np.asarray(mmp[f"elec_2_{gauge_origin}"])
            if mmp["elec_2_origin"].shape != (9, nb, nb):
                raise ValueError("multipoles/elec_2_origin is expected to "
                                 "have shape " + str((9, nb, nb)) + " not "
                                 + str(mmp["elec_2"].shape))
            opprov.electric_quadrupole = get_integral_elquad
        magm = data.get("magnetic_moments", {})
        if "mag_1_origin" in magm:
            def get_integral_magdip(gauge_origin):
                return np.asarray(magm[f"mag_1_{gauge_origin}"])
            if magm["mag_1_origin"].shape != (3, nb, nb):
                raise ValueError("magnetic_moments/mag_1_origin is expected to have"
                                 " shape " + str((3, nb, nb)) + " not "
                                 + str(magm["mag_1_origin"].shape))
            opprov.magnetic_dipole = get_integral_magdip
        derivs = data.get("derivatives", {})
        if "elec_vel_1" in derivs:
            if derivs["elec_vel_1"].shape != (3, nb, nb):
                raise ValueError("derivatives/elec_vel_1 is expected to have shape "
                                 + str((3, nb, nb)) + " not "
                                 + str(derivs["elec_vel_1"].shape))
            opprov.electric_dipole_velocity = np.asarray(derivs["elec_vel_1"])
        if "overlap" in data:
            if data["overlap"].shape != (nb, nb):
                raise ValueError("overlap is expected to have shape "
                                 + str((nb, nb)) + " not "
                                 + str(data["overlap"].shape))
            opprov.overlap = np.asarray(data["overlap"])

        self.operator_integral_provider = opprov

    #
    # Required keys
    #
    def get_restricted(self):
        return get_scalar_value(self.data, "restricted")

    def get_conv_tol(self):
        if "conv_tol" in self.data:
            return get_scalar_value(self.data, "conv_tol")
        # The old name was "threshold"
        return get_scalar_value(self.data, "threshold")

    def fill_occupation_f(self, out):
        out[:] = self.data["occupation_f"]

    def fill_orbcoeff_fb(self, out):
        out[:] = self.data["orbcoeff_fb"]

    def fill_orben_f(self, out):
        out[:] = self.data["orben_f"]

    def fill_fock_ff(self, slices, out):
        out[:] = self.data["fock_ff"][slices]

    def fill_eri_ffff(self, slices, out):
        out[:] = self.data["eri_ffff"][slices]

    def fill_eri_phys_asym_ffff(self, slices, out):
        # Only required if eri_ffff not provided
        out[:] = self.data["eri_phys_asym_ffff"][slices]

    #
    # Recommended keys
    #
    def get_backend(self):
        return self.__backend

    def get_energy_scf(self):
        return get_scalar_value(self.data, "energy_scf", 0.0)

    def get_nuclear_multipole(self, order: int, gauge_origin: str) -> np.ndarray:
        key = f"multipoles/nuclear_{order}"
        if order == 0:
            nuc_multipole = get_scalar_value(self.data, key, default=None)
        elif order == 2:
            key_quad = key + f"_{gauge_origin}"
            nuc_multipole = get_array_value(self.data, key_quad, default=None)
        else:
            nuc_multipole = get_array_value(self.data, key, default=None)
        if nuc_multipole is None:
            raise NotImplementedError(f"Nuclear multipole with order {order} is "
                                      "not available.")
        if order == 0:  # The function interface needs an np.array on return
            nuc_multipole = np.array([nuc_multipole])
        return nuc_multipole

    def get_spin_multiplicity(self):
        if "spin_multiplicity" in self.data:
            return get_scalar_value(self.data, "spin_multiplicity")
        elif not self.get_restricted():
            return 0
        else:
            noa = self.get_n_orbs_alpha()
            na = int(np.sum(self.data["occupation_f"][:noa]))
            nb = int(np.sum(self.data["occupation_f"][noa:]))
            return na - nb + 1

    #
    # Deduced keys
    #
    def get_n_orbs_alpha(self):
        return self.data["orbcoeff_fb"].shape[0] // 2

    def get_n_bas(self):
        return self.data["orbcoeff_fb"].shape[1]

    def has_eri_phys_asym_ffff_inner(self):
        return "eri_phys_asym_ffff" in self.data


class DictHfProvider(DataHfProvider):
    def __init__(self, *args, **kwargs):
        from warnings import warn

        super().__init__(*args, **kwargs)
        warn(DeprecationWarning("DictHfProvider is deprecated, "
                                "use DataHfProvider"))



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .Symmetry import Symmetry

import libadcc


class Tensor(libadcc.Tensor):
    def __init__(self, sym_or_mo, space=None,
                 permutations=None, spin_block_maps=None,
                 spin_blocks_forbidden=None):
        """Construct an uninitialised Tensor from an :class:`MoSpaces` or
        a :class:`Symmetry` object.

        More information about the last four, symmetry-related parameters
        see the documentation of the :class:`Symmetry` object.

        Parameters
        ----------
        sym_or_mo
            Symmetry or MoSpaces object
        spaces : str, optional
            Space of the tensor, can be None if the first argument is
            a :class:`Symmetry` object.
        permutations : list, optional
            List of permutational symmetries of the Tensor.
        spin_block_maps : list, optional
            List of mappings between spin blocks
        spin_blocks_forbidden : list, optional
            List of forbidden (i.e. forced-to-zero) spin blocks.

        Notes
        -----
        An :class:`MoSpaces` object is contained in many datastructures
        of adcc, including the :class:`AdcMatrix`, the :class:`LazyMp`,
        the :class:`ReferenceState` and any solver or ADC results state.

        Examples
        --------
        Construct a symmetric tensor in the "o1o1" (occupied-occupied) spaces:

        >>> Tensor(mospaces, "o1o1", permutations=["ij", "ji"])

        Construct an anti-symmetric tensor in the "v1v1" spaces:

        >>> Tensor(mospaces, "v1v1", permutations=["ab", "-ba"])

        Construct a tensor in "o1v1", which maps the alpha-alpha block
        anti-symmetrically to the beta-beta block and which has the
        other spin blocks set to zero:

        >>> Tensor(mospaces, "o1v1", spin_block_maps=[("aa", "bb", -1)],
        ...        spin_blocks_forbidden=["ab", "ba"])

        """
        if not isinstance(sym_or_mo, (libadcc.MoSpaces, libadcc.Symmetry)):
            raise TypeError("The first argument needs to be a Symmetry or an "
                            "MoSpaces object.")
        if not isinstance(sym_or_mo, libadcc.Symmetry):
            if space is None:
                raise ValueError("If the first argument to Tensor is no "
                                 "Symmetry object, the second argument (spaces)"
                                 "needs to be given")
            sym_or_mo = Symmetry(sym_or_mo, space, permutations,
                                 spin_block_maps, spin_blocks_forbidden)

        if space is not None:
            if sym_or_mo.space != space:
                raise ValueError("Value passed to space needs to agree with "
                                 "space value from Symmetry object.")

        super().__init__(sym_or_mo)


def _tensor_select_below_absmax(tensor, tolerance):
    """
    Select the absolute maximal values in the tensor,
    which are below the given tolerance.
    """
    n = min(10, tensor.size)
    res = []
    while n <= tensor.size:
        res = tensor.select_n_absmax(n)
        minampl = min(abs(r[1]) for r in res)
        if minampl < tolerance:
            break
        else:
            n = max(n + 1, min(tensor.size, 2 * n))
    return [r for r in res if abs(r[1]) >= tolerance]


Tensor.select_below_absmax = _tensor_select_below_absmax



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import subprocess
import numpy as np

N = 3
a = np.random.randn(N, N, N, N)

a_sym_01 = 0.5 * (a + a.transpose((1, 0, 2, 3)))
a_asym_01 = 0.5 * (a - a.transpose((1, 0, 2, 3)))
a_sym_01_23 = 0.5 * (a + a.transpose((1, 0, 3, 2)))
a_asym_01_23 = 0.5 * (a - a.transpose((1, 0, 3, 2)))
a_sym_012 = 1 / 6 * (
    a + a.transpose((0, 2, 1, 3))
    + a.transpose((1, 0, 2, 3)) + a.transpose((1, 2, 0, 3))
    + a.transpose((2, 1, 0, 3)) + a.transpose((2, 0, 1, 3))
)
a_asym_012 = 1 / 6 * (
    a - a.transpose((0, 2, 1, 3))
    - a.transpose((1, 0, 2, 3)) + a.transpose((1, 2, 0, 3))
    - a.transpose((2, 1, 0, 3)) + a.transpose((2, 0, 1, 3))
)

#
#
#

HEADER = """
//
// Copyright (C) 2018 by the adcc authors
//
// This file is part of adcc.
//
// adcc is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// adcc is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with adcc. If not, see <http://www.gnu.org/licenses/>.
//

#include "TensorTestData.hh"

"""


def npprint(arr):
    fmt = '{:.18e}'
    str_arr = [fmt.format(x) for x in arr]
    return ",".join(str_arr)


with open("./TensorTestData.cc", "w") as f:
    f.write(HEADER)
    f.write("namespace libadcc {\nnamespace tests {\n\n")
    f.write("size_t TensorTestData::N = " + str(N) + ";")
    f.write("std::vector<double> TensorTestData::a = {"
            + npprint(a.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_sym_01 = {"
            + npprint(a_sym_01.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_asym_01 = {"
            + npprint(a_asym_01.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_sym_01_23 = {"
            + npprint(a_sym_01_23.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_asym_01_23 = {"
            + npprint(a_asym_01_23.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_sym_012 = {"
            + npprint(a_sym_012.ravel()) + "};")
    f.write("std::vector<double> TensorTestData::a_asym_012 = {"
            + npprint(a_asym_012.ravel()) + "};")
    f.write("\n}\n}")


subprocess.check_call(["clang-format", "-i", "./TensorTestData.cc"])



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys
import warnings

from libadcc import ReferenceState

from . import solver
from .guess import (guesses_any, guesses_singlet, guesses_spin_flip,
                    guesses_triplet)
from .LazyMp import LazyMp
from .AdcMatrix import AdcMatrix, AdcMatrixlike, AdcExtraTerm
from .AdcMethod import AdcMethod
from .exceptions import InputError
from .ExcitedStates import ExcitedStates
from .ReferenceState import ReferenceState as adcc_ReferenceState
from .solver.lanczos import lanczos
from .solver.davidson import jacobi_davidson
from .solver.explicit_symmetrisation import (IndexSpinSymmetrisation,
                                             IndexSymmetrisation)

__all__ = ["run_adc"]


def run_adc(data_or_matrix, n_states=None, kind="any", conv_tol=None,
            eigensolver=None, guesses=None, n_guesses=None,
            n_guesses_doubles=None, output=sys.stdout, core_orbitals=None,
            frozen_core=None, frozen_virtual=None, method=None,
            n_singlets=None, n_triplets=None, n_spin_flip=None,
            environment=None, **solverargs):
    """Run an ADC calculation.

    Main entry point to run an ADC calculation. The reference to build the ADC
    calculation upon is supplied using the `data_or_matrix` argument.
    `adcc` is pretty flexible here. Possible options include:

        a. Hartree-Fock data from a host program, e.g. a molsturm SCF
           state, a pyscf SCF object or any class implementing the
           :py:class:`adcc.HartreeFockProvider` interface. From this data all
           objects mentioned in (b) to (d) will be implicitly created and will
           become available in the returned state.
        b. A :py:class:`adcc.ReferenceState` object
        c. A :py:class:`adcc.LazyMp` object
        d. A :py:class:`adcc.AdcMatrix` object

    Parameters
    ----------
    data_or_matrix
        Data containing the SCF reference
    n_states : int, optional
    kind : str, optional
    n_singlets : int, optional
    n_triplets : int, optional
    n_spin_flip : int, optional
        Specify the number and kind of states to be computed. Possible values
        for kind are "singlet", "triplet", "spin_flip" and "any", which is
        the default. For unrestricted references clamping spin-pure
        singlets/triplets is currently not possible and kind has to remain as
        "any". For restricted references `kind="singlets"` or `kind="triplets"`
        may be employed to enforce a particular excited states manifold.
        Specifying `n_singlets` is equivalent to setting `kind="singlet"` and
        `n_states=5`. Similarly for `n_triplets` and `n_spin_flip`.
        `n_spin_flip` is only valid for unrestricted references.

    conv_tol : float, optional
        Convergence tolerance to employ in the iterative solver for obtaining
        the ADC vectors (default: `1e-6` or 10 * SCF tolerance,
        whatever is larger)

    eigensolver : str, optional
        The eigensolver algorithm to use.

    n_guesses : int, optional
        Total number of guesses to compute. By default only guesses derived from
        the singles block of the ADC matrix are employed. See
        `n_guesses_doubles` for alternatives. If no number is given here
        `n_guesses = min(4, 2 * number of excited states to compute)`
        or a smaller number if the number of excitation is estimated to be less
        than the outcome of above formula.

    n_guesses_doubles : int, optional
        Number of guesses to derive from the doubles block. By default none
        unless n_guesses as explicitly given or automatically determined is
        larger than the number of singles guesses, which can be possibly found.

    guesses : list, optional
        Provide the guess vectors to be employed for the ADC run. Takes
        preference over `n_guesses` and `n_guesses_doubles`, such that these
        parameters are ignored.

    output : stream, optional
        Python stream to which output will be written. If `None` all output
        is disabled.

    core_orbitals : int or list or tuple, optional
        The orbitals to be put into the core-occupied space. For ways to
        define the core orbitals see the description in
        :py:class:`adcc.ReferenceState`.
        Required if core-valence separation is applied and the input data is
        given as data from the host program (i.e. option (a) discussed above)

    frozen_core : int or list or tuple, optional
        The orbitals to select as frozen core orbitals (i.e. inactive occupied
        orbitals for both the MP and ADC methods performed). For ways to define
        these see the description in :py:class:`adcc.ReferenceState`.

    frozen_virtual : int or list or tuple, optional
        The orbitals to select as frozen virtual orbitals (i.e. inactive
        virtuals for both the MP and ADC methods performed). For ways to define
        these see the description in :py:class:`adcc.ReferenceState`.

    environment : bool or list or dict, optional
        The keywords to specify how coupling to an environment model,
        e.g. PE, is treated. For details see :ref:`environment`.

    Other parameters
    ----------------
    max_subspace : int, optional
        Maximal subspace size
    max_iter : int, optional
        Maximal number of iterations
    max_subspace_iter : int, optional
        Maximum number of iterations for diagonalizing the subspace matrix
        during the Davidson solver procedure

    Returns
    -------
    ExcitedStates
        An :class:`adcc.ExcitedStates` object containing the
        :class:`adcc.AdcMatrix`, the :class:`adcc.LazyMp` ground state and the
        :class:`adcc.ReferenceState` as well as computed eigenpairs.

    Examples
    --------

    Run an ADC(2) calculation on top of a `pyscf` RHF reference of
    hydrogen flouride.

    >>> from pyscf import gto, scf
    ... mol = gto.mole.M(atom="H 0 0 0; F 0 0 1.1", basis="sto-3g")
    ... mf = scf.RHF(mol)
    ... mf.conv_tol_grad = 1e-8
    ... mf.kernel()
    ...
    ... state = adcc.run_adc(mf, method="adc2", n_singlets=3)

    The same thing can also be achieved using the `adcc.adcN` family of
    short-hands (see e.g. :py:func:`adcc.adc2`, :py:func:`adcc.cvs_adc2x`):

    >>> state = adcc.adc2(mf, n_singlets=3)

    Run a CVS-ADC(3) calculation of O2 with one core-occupied orbital

    >>> from pyscf import gto, scf
    ... mol = gto.mole.M(atom="O 0 0 0; O 0 0 1.2", basis="sto-3g")
    ... mf = scf.RHF(mol)
    ... mf.conv_tol_grad = 1e-8
    ... mf.kernel()
    ...
    ... state = adcc.cvs_adc3(mf, core_orbitals=1, n_singlets=3)
    """
    matrix = construct_adcmatrix(
        data_or_matrix, core_orbitals=core_orbitals, frozen_core=frozen_core,
        frozen_virtual=frozen_virtual, method=method)

    n_states, kind = validate_state_parameters(
        matrix.reference_state, n_states=n_states, n_singlets=n_singlets,
        n_triplets=n_triplets, n_spin_flip=n_spin_flip, kind=kind)

    # Determine spin change during excitation. If guesses is not None,
    # i.e. user-provided, we cannot guarantee for obtaining a particular
    # spin_change in case of a spin_flip calculation.
    spin_change = None
    if kind == "spin_flip" and guesses is None:
        spin_change = -1

    # Select solver to run
    if eigensolver is None:
        eigensolver = "davidson"

    # Setup environment coupling terms and energy corrections
    ret = setup_environment(matrix, environment)
    env_matrix_term, env_energy_corrections = ret
    # add terms to matrix
    if env_matrix_term:
        matrix += env_matrix_term

    diagres = diagonalise_adcmatrix(
        matrix, n_states, kind, guesses=guesses, n_guesses=n_guesses,
        n_guesses_doubles=n_guesses_doubles, conv_tol=conv_tol, output=output,
        eigensolver=eigensolver, **solverargs)
    exstates = ExcitedStates(diagres)
    exstates.kind = kind
    exstates.spin_change = spin_change

    # add environment corrections to excited states
    exstates += env_energy_corrections
    return exstates


#
# Individual steps
#
def construct_adcmatrix(data_or_matrix, core_orbitals=None, frozen_core=None,
                        frozen_virtual=None, method=None):
    """
    Use the provided data or AdcMatrix object to check consistency of the
    other passed parameters and construct the AdcMatrix object representing
    the problem to be solved.
    Internal function called from run_adc.
    """
    if not isinstance(data_or_matrix, AdcMatrixlike) and method is None:
        raise InputError("method needs to be explicitly provided unless "
                         "data_or_matrix is an AdcMatrixlike.")
    if method is not None and not isinstance(method, AdcMethod):
        try:
            method = AdcMethod(method)
        except ValueError as e:
            raise InputError(str(e))  # In case the method is unknown

    if not isinstance(data_or_matrix, (ReferenceState, AdcMatrixlike, LazyMp)):
        if method.is_core_valence_separated and core_orbitals is None:
            raise InputError("If core-valence separation approximation is "
                             "applied then the number of core orbitals needs "
                             "to be specified via the parameter "
                             "core_orbitals.")
        try:
            refstate = adcc_ReferenceState(data_or_matrix,
                                           core_orbitals=core_orbitals,
                                           frozen_core=frozen_core,
                                           frozen_virtual=frozen_virtual)
        except ValueError as e:
            raise InputError(str(e))  # In case of an issue with the spaces
        data_or_matrix = refstate
    elif core_orbitals is not None:
        mospaces = data_or_matrix.mospaces
        warnings.warn("Ignored core_orbitals parameter because data_or_matrix"
                      " is a ReferenceState, a LazyMp or an AdcMatrixlike object "
                      " (which has a value of core_orbitals={})."
                      "".format(mospaces.n_orbs_alpha("o2")))
    elif frozen_core is not None:
        mospaces = data_or_matrix.mospaces
        warnings.warn("Ignored frozen_core parameter because data_or_matrix"
                      " is a ReferenceState, a LazyMp or an AdcMatrixlike object "
                      " (which has a value of frozen_core={})."
                      "".format(mospaces.n_orbs_alpha("o3")))
    elif frozen_virtual is not None:
        mospaces = data_or_matrix.mospaces
        warnings.warn("Ignored frozen_virtual parameter because data_or_matrix"
                      " is a ReferenceState, a LazyMp or an AdcMatrixlike object "
                      " (which has a value of frozen_virtual={})."
                      "".format(mospaces.n_orbs_alpha("v2")))

    # Make AdcMatrix (if not done)
    if isinstance(data_or_matrix, (ReferenceState, LazyMp)):
        try:
            return AdcMatrix(method, data_or_matrix)
        except ValueError as e:
            # In case of an issue with CVS <-> chosen spaces
            raise InputError(str(e))
    elif method is not None and method != data_or_matrix.method:
        warnings.warn("Ignored method parameter because data_or_matrix is an"
                      " AdcMatrixlike, which implicitly sets the method")
    if isinstance(data_or_matrix, AdcMatrixlike):
        return data_or_matrix


def validate_state_parameters(reference_state, n_states=None, n_singlets=None,
                              n_triplets=None, n_spin_flip=None, kind="any"):
    """
    Check the passed state parameters for consistency with itself and with
    the passed reference and normalise them. In the end return the number of
    states and the corresponding kind parameter selected.
    Internal function called from run_adc.
    """
    if sum(nst is not None for nst in [n_states, n_singlets,
                                       n_triplets, n_spin_flip]) > 1:
        raise InputError("One may only specify one out of n_states, "
                         "n_singlets, n_triplets and n_spin_flip")

    if n_singlets is not None:
        if not reference_state.restricted:
            raise InputError("The n_singlets parameter may only be employed "
                             "for restricted references")
        if kind not in ["singlet", "any"]:
            raise InputError(f"Kind parameter {kind} not compatible "
                             "with n_singlets > 0")
        kind = "singlet"
        n_states = n_singlets
    if n_triplets is not None:
        if not reference_state.restricted:
            raise InputError("The n_triplets parameter may only be employed "
                             "for restricted references")
        if kind not in ["triplet", "any"]:
            raise InputError(f"Kind parameter {kind} not compatible "
                             "with n_triplets > 0")
        kind = "triplet"
        n_states = n_triplets
    if n_spin_flip is not None:
        if reference_state.restricted:
            raise InputError("The n_spin_flip parameter may only be employed "
                             "for unrestricted references")
        if kind not in ["spin_flip", "any"]:
            raise InputError(f"Kind parameter {kind} not compatible "
                             "with n_spin_flip > 0")
        kind = "spin_flip"
        n_states = n_spin_flip

    # Check if there are states to be computed
    if n_states is None or n_states == 0:
        raise InputError("No excited states to be computed. Specify at least "
                         "one of n_states, n_singlets, n_triplets, "
                         "or n_spin_flip")
    if n_states < 0:
        raise InputError("n_states needs to be positive")

    if kind not in ["any", "spin_flip", "singlet", "triplet"]:
        raise InputError("The kind parameter may only take the values 'any', "
                         "'singlet', 'triplet' or 'spin_flip'")
    if kind in ["singlet", "triplet"] and not reference_state.restricted:
        raise InputError("kind==singlet and kind==triplet are only valid for "
                         "ADC calculations in combination with a restricted "
                         "ground state.")
    if kind in ["spin_flip"] and reference_state.restricted:
        raise InputError("kind==spin_flip is only valid for "
                         "ADC calculations in combination with an unrestricted "
                         "ground state.")
    return n_states, kind


def diagonalise_adcmatrix(matrix, n_states, kind, eigensolver="davidson",
                          guesses=None, n_guesses=None, n_guesses_doubles=None,
                          conv_tol=None, output=sys.stdout, **solverargs):
    """
    This function seeks appropriate guesses and afterwards proceeds to
    diagonalise the ADC matrix using the specified eigensolver.
    Internal function called from run_adc.
    """
    reference_state = matrix.reference_state

    # Determine default ADC convergence tolerance
    if conv_tol is None:
        conv_tol = max(10 * reference_state.conv_tol, 1e-6)
    if reference_state.conv_tol > conv_tol:
        raise InputError(
            "Convergence tolerance of SCF results "
            f"(== {reference_state.conv_tol}) needs to be lower than ADC "
            f"convergence tolerance parameter conv_tol (== {conv_tol})."
        )

    # Determine explicit_symmetrisation
    explicit_symmetrisation = IndexSymmetrisation
    if kind in ["singlet", "triplet"]:
        explicit_symmetrisation = IndexSpinSymmetrisation(
            matrix, enforce_spin_kind=kind
        )

    # Set some solver-specific parameters
    if eigensolver == "davidson":
        n_guesses_per_state = 2
        callback = setup_solver_printing(
            "Jacobi-Davidson", matrix, kind, solver.davidson.default_print,
            output=output)
        run_eigensolver = jacobi_davidson
    elif eigensolver == "lanczos":
        n_guesses_per_state = 1
        callback = setup_solver_printing(
            "Lanczos", matrix, kind, solver.lanczos.default_print,
            output=output)
        run_eigensolver = lanczos
    else:
        raise InputError(f"Solver {eigensolver} unknown, try 'davidson'.")

    # Obtain or check guesses
    if guesses is None:
        if n_guesses is None:
            # restrict to the number of available singles guesses if no doubles
            # are available
            n_guesses = estimate_n_guesses(
                matrix=matrix, n_states=n_states,
                singles_only=("pphh" not in matrix.axis_blocks),
                n_guesses_per_state=n_guesses_per_state
            )
        guesses = obtain_guesses_by_inspection(matrix, n_guesses, kind,
                                               n_guesses_doubles)
    else:
        if len(guesses) < n_states:
            raise InputError("Less guesses provided via guesses (== {}) "
                             "than states to be computed (== {})"
                             "".format(len(guesses), n_states))
        if n_guesses is not None:
            warnings.warn("Ignoring n_guesses parameter, since guesses are "
                          "explicitly provided.")
        if n_guesses_doubles is not None:
            warnings.warn("Ignoring n_guesses_doubles parameter, since guesses "
                          "are explicitly provided.")

    solverargs.setdefault("which", "SA")
    return run_eigensolver(matrix, guesses, n_ep=n_states, conv_tol=conv_tol,
                           callback=callback,
                           explicit_symmetrisation=explicit_symmetrisation,
                           **solverargs)


def estimate_n_guesses(matrix, n_states, singles_only=True,
                       n_guesses_per_state=2):
    """
    Implementation of a basic heuristic to find a good number of guess
    vectors to be searched for using the find_guesses function.
    Internal function called from run_adc.

    matrix             ADC matrix
    n_states           Number of states to be computed
    singles_only       Try to stay withing the singles excitation space
                       with the number of guess vectors.
    n_guesses_per_state  Number of guesses to search for for each state
    """
    # Try to use at least 4 or twice the number of states
    # to be computed as guesses
    n_guesses = n_guesses_per_state * max(2, n_states)

    if singles_only:
        # Compute the maximal number of sensible singles block guesses.
        # This is roughly the number of occupied alpha orbitals
        # times the number of virtual alpha orbitals
        #
        # If the system is core valence separated, then only the
        # core electrons count as "occupied".
        mospaces = matrix.mospaces
        sp_occ = "o2" if matrix.is_core_valence_separated else "o1"
        n_virt_a = mospaces.n_orbs_alpha("v1")
        n_occ_a = mospaces.n_orbs_alpha(sp_occ)
        n_guesses = min(n_guesses, n_occ_a * n_virt_a)

    # Adjust if we overshoot the maximal number of sensible singles block
    # guesses, but make sure we get at least n_states guesses
    return max(n_states, n_guesses)


def obtain_guesses_by_inspection(matrix, n_guesses, kind, n_guesses_doubles=None):
    """
    Obtain guesses by inspecting the diagonal matrix elements.
    If n_guesses_doubles is not None, this is number is always adhered to.
    Otherwise the number of doubles guesses is adjusted to fill up whatever
    the singles guesses cannot provide to reach n_guesses.
    Internal function called from run_adc.
    """
    if n_guesses_doubles is not None and n_guesses_doubles > 0 \
       and "pphh" not in matrix.axis_blocks:
        raise InputError("n_guesses_doubles > 0 is only sensible if the ADC "
                         "method has a doubles block (i.e. it is *not* ADC(0), "
                         "ADC(1) or a variant thereof.")

    # Determine guess function
    guess_function = {"any": guesses_any, "singlet": guesses_singlet,
                      "triplet": guesses_triplet,
                      "spin_flip": guesses_spin_flip}[kind]

    # Determine number of singles guesses to request
    n_guess_singles = n_guesses
    if n_guesses_doubles is not None:
        n_guess_singles = n_guesses - n_guesses_doubles
    singles_guesses = guess_function(matrix, n_guess_singles, block="ph")

    doubles_guesses = []
    if "pphh" in matrix.axis_blocks:
        # Determine number of doubles guesses to request if not
        # explicitly specified
        if n_guesses_doubles is None:
            n_guesses_doubles = n_guesses - len(singles_guesses)
        if n_guesses_doubles > 0:
            doubles_guesses = guess_function(matrix, n_guesses_doubles,
                                             block="pphh")

    total_guesses = singles_guesses + doubles_guesses
    if len(total_guesses) < n_guesses:
        raise InputError("Less guesses found than requested: {} found, "
                         "{} requested".format(len(total_guesses), n_guesses))
    return total_guesses


def setup_solver_printing(solmethod_name, matrix, kind, default_print,
                          output=None):
    """
    Setup default printing for solvers. Internal function called from run_adc.
    """
    kstr = " "
    if kind != "any":
        kstr = " " + kind
    method_name = f"{matrix}"
    if hasattr(matrix, "method"):
        method_name = matrix.method.name

    if output is not None:
        print(f"Starting {method_name}{kstr} {solmethod_name} ...",
              file=output)

        def inner_callback(state, identifier):
            default_print(state, identifier, output)
        return inner_callback


def setup_environment(matrix, environment):
    """
    Setup environment matrix terms and/or energy corrections.
    Internal function called from run_adc.
    """
    valid_envs = ["ptss", "ptlr", "linear_response"]
    hf = matrix.reference_state
    if hf.environment and environment is None:
        raise InputError(
            "Environment found in reference state, but no environment"
            " configuration specified. Please select from the following"
            f" schemes: {valid_envs} or set to False."
        )
    elif environment and not hf.environment:
        raise InputError(
            "Environment specified, but no environment"
            " was found in reference state."
        )
    elif not hf.environment:
        environment = {}

    convertor = {
        bool: lambda value: {"ptss": True, "ptlr": True} if value else {},
        list: lambda value: {k: True for k in value},
        str: lambda value: {value: True},
        dict: lambda value: value,
    }
    conversion = convertor.get(type(environment), None)
    if conversion is None:
        raise TypeError("Cannot convert environment parameter of type"
                        f"'{type(environment)}' to dict.")
    environment = conversion(environment)

    if any(env not in valid_envs for env in environment):
        raise InputError("Invalid key specified for environment."
                         f" Valid keys are '{valid_envs}'.")

    env_matrix_term = None
    energy_corrections = []

    forbidden_combinations = [
        ["ptlr", "linear_response"],
    ]
    for fbc in forbidden_combinations:
        if all(environment.get(k, False) for k in fbc):
            raise InputError("Combination of environment schemes"
                             f" '{fbc}' not allowed. Check the"
                             " adcc documentation for more details.")

    for pt in ["ptss", "ptlr"]:
        if not environment.get(pt, False):
            continue
        hf_corr = hf.excitation_energy_corrections
        eec_key = f"{hf.environment}_{pt}_correction"
        if eec_key not in hf_corr:
            raise ValueError(f"{pt} correction requested, but could not find"
                             f" the needed function {eec_key} in"
                             f" reference state from backend {hf.backend}.")
        energy_corrections.append(hf_corr[eec_key])
    if environment.get("linear_response", False):
        from adcc.adc_pp import environment as adcpp_env
        block_key = f"block_ph_ph_0_{hf.environment}"
        if not hasattr(adcpp_env, block_key):
            raise NotImplementedError("Matrix term for linear response coupling"
                                      f" with solvent {hf.environment}"
                                      " not implemented.")
        block_fun = getattr(adcpp_env, block_key)
        env_matrix_term = AdcExtraTerm(matrix, {'ph_ph': block_fun})

    return env_matrix_term, energy_corrections



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2022 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from collections import namedtuple
from adcc import block as b
from adcc.functions import einsum
from adcc.AmplitudeVector import AmplitudeVector


__all__ = ["block"]


#
# Dispatch routine
#

"""
`apply` is a function mapping an AmplitudeVector to the contribution of this
block to the result of applying the ISR matrix.
"""
IsrBlock = namedtuple("IsrBlock", ["apply"])


def block(ground_state, operator, spaces, order, variant=None):
    """
    Gets ground state, one-particle matrix elements associated
    with a one-particle operator, spaces (ph, pphh and so on)
    and the perturbation theory order for the block,
    variant is "cvs" or sth like that.

    The matrix-vector product was derived up to second order
    using the original equations from
    J. Schirmer and A. B. Trofimov, J. Chem. Phys. 120, 11449–11464 (2004).
    """
    if isinstance(variant, str):
        variant = [variant]
    elif variant is None:
        variant = []

    if ground_state.has_core_occupied_space and "cvs" not in variant:
        raise ValueError("Cannot run a general (non-core-valence approximated) "
                         "ADC method on top of a ground state with a "
                         "core-valence separation.")
    if not ground_state.has_core_occupied_space and "cvs" in variant:
        raise ValueError("Cannot run a core-valence approximated ADC method on "
                         "top of a ground state without a "
                         "core-valence separation.")

    fn = "_".join(["block"] + variant + spaces + [str(order)])

    if fn not in globals():
        raise ValueError("Could not dispatch: "
                         f"spaces={spaces} order={order} variant={variant}. "
                         "Probably the B-matrix is not implemented for the "
                         "requested method.")
    return globals()[fn](ground_state, operator)


#
# 0th order main
#
def block_ph_ph_0(ground_state, op):
    def apply(ampl):
        return AmplitudeVector(ph=(
            + 1.0 * einsum('ic,ac->ia', ampl.ph, op.vv)
            - 1.0 * einsum('ka,ki->ia', ampl.ph, op.oo)
        ))
    return IsrBlock(apply)


def block_pphh_pphh_0(ground_state, op):
    def apply(ampl):
        return AmplitudeVector(pphh=0.5 * (
            (
                + 2.0 * einsum('ijcb,ac->ijab', ampl.pphh, op.vv)
                - 2.0 * einsum('ijca,bc->ijab', ampl.pphh, op.vv)
            ).antisymmetrise(2, 3)
            + (
                - 2.0 * einsum('kjab,ki->ijab', ampl.pphh, op.oo)
                + 2.0 * einsum('kiab,kj->ijab', ampl.pphh, op.oo)
            ).antisymmetrise(0, 1)
        ))
    return IsrBlock(apply)


#
# 0th order coupling
#
def block_ph_pphh_0(ground_state, op):
    def apply(ampl):
        return AmplitudeVector(ph=0.5 * (
            - 2.0 * einsum('ilad,ld->ia', ampl.pphh, op.ov)
            + 2.0 * einsum('ilca,lc->ia', ampl.pphh, op.ov)
        ))
    return IsrBlock(apply)


def block_pphh_ph_0(ground_state, op):
    def apply(ampl):
        return AmplitudeVector(pphh=0.5 * (
            (
                - 1.0 * einsum('ia,bj->ijab', ampl.ph, op.vo)
                + 1.0 * einsum('ja,bi->ijab', ampl.ph, op.vo)
                + 1.0 * einsum('ib,aj->ijab', ampl.ph, op.vo)
                - 1.0 * einsum('jb,ai->ijab', ampl.ph, op.vo)
            ).antisymmetrise(0, 1).antisymmetrise(2, 3)
        ))
    return IsrBlock(apply)


#
# 1st order main
#
block_ph_ph_1 = block_ph_ph_0


#
# 1st order coupling
#
def block_ph_pphh_1(ground_state, op):
    t2 = ground_state.t2(b.oovv)

    def apply(ampl):
        return AmplitudeVector(ph=0.5 * (
            # zeroth order
            - 2.0 * einsum('ilad,ld->ia', ampl.pphh, op.ov)
            + 2.0 * einsum('ilca,lc->ia', ampl.pphh, op.ov)
            # first order
            + 2.0 * einsum('ilad,lndf,fn->ia', ampl.pphh, t2, op.vo)
            - 2.0 * einsum('ilca,lncf,fn->ia', ampl.pphh, t2, op.vo)
            - 2.0 * einsum('klad,kled,ei->ia', ampl.pphh, t2, op.vo)
            - 2.0 * einsum('ilcd,nlcd,an->ia', ampl.pphh, t2, op.vo)
        ))
    return IsrBlock(apply)


def block_pphh_ph_1(ground_state, op):
    t2 = ground_state.t2(b.oovv)

    def apply(ampl):
        return AmplitudeVector(pphh=0.5 * (
            (
                # zeroth order
                - 1.0 * einsum('ia,bj->ijab', ampl.ph, op.vo)
                + 1.0 * einsum('ja,bi->ijab', ampl.ph, op.vo)
                + 1.0 * einsum('ib,aj->ijab', ampl.ph, op.vo)
                - 1.0 * einsum('jb,ai->ijab', ampl.ph, op.vo)
                # first order
                + 1.0 * einsum('ia,jnbf,nf->ijab', ampl.ph, t2, op.ov)
                - 1.0 * einsum('ja,inbf,nf->ijab', ampl.ph, t2, op.ov)
                - 1.0 * einsum('ib,jnaf,nf->ijab', ampl.ph, t2, op.ov)
                + 1.0 * einsum('jb,inaf,nf->ijab', ampl.ph, t2, op.ov)
            ).antisymmetrise(0, 1).antisymmetrise(2, 3)
            + (
                - 1.0 * einsum('ka,ijeb,ke->ijab', ampl.ph, t2, op.ov)
                + 1.0 * einsum('kb,ijea,ke->ijab', ampl.ph, t2, op.ov)
            ).antisymmetrise(2, 3)
            + (
                - 1.0 * einsum('ic,njab,nc->ijab', ampl.ph, t2, op.ov)
                + 1.0 * einsum('jc,niab,nc->ijab', ampl.ph, t2, op.ov)
            ).antisymmetrise(0, 1)
        ))
    return IsrBlock(apply)


#
# 2nd order main
#
def block_ph_ph_2(ground_state, op):
    p0 = ground_state.mp2_diffdm
    t2 = ground_state.t2(b.oovv)

    def apply(ampl):
        return AmplitudeVector(ph=(
            # 0th order
            + 1.0 * einsum('ic,ac->ia', ampl.ph, op.vv)
            - 1.0 * einsum('ka,ki->ia', ampl.ph, op.oo)
            # 2nd order
            # (2,1)
            - 1.0 * einsum('ic,jc,aj->ia', ampl.ph, p0.ov, op.vo)
            - 1.0 * einsum('ka,kb,bi->ia', ampl.ph, p0.ov, op.vo)
            - 1.0 * einsum('ic,ja,jc->ia', ampl.ph, p0.ov, op.ov)  # h.c.
            - 1.0 * einsum('ka,ib,kb->ia', ampl.ph, p0.ov, op.ov)  # h.c.
            # (2,2)
            - 0.25 * einsum('ic,mnef,mnaf,ec->ia', ampl.ph, t2, t2, op.vv)
            - 0.25 * einsum('ic,mnef,mncf,ae->ia', ampl.ph, t2, t2, op.vv)  # h.c.
            # (2,3)
            - 0.5 * einsum('ic,mnce,mnaf,ef->ia', ampl.ph, t2, t2, op.vv)
            + 1.0 * einsum('ic,mncf,jnaf,jm->ia', ampl.ph, t2, t2, op.oo)
            # (2,4)
            + 0.25 * einsum('ka,mnef,inef,km->ia', ampl.ph, t2, t2, op.oo)
            + 0.25 * einsum('ka,mnef,knef,mi->ia', ampl.ph, t2, t2, op.oo)  # h.c.
            # (2,5)
            - 1.0 * einsum('ka,knef,indf,ed->ia', ampl.ph, t2, t2, op.vv)
            + 0.5 * einsum('ka,knef,imef,mn->ia', ampl.ph, t2, t2, op.oo)
            # (2,6)
            + 0.5 * einsum('kc,knef,inaf,ec->ia', ampl.ph, t2, t2, op.vv)
            - 0.5 * einsum('kc,mncf,inaf,km->ia', ampl.ph, t2, t2, op.oo)
            + 0.5 * einsum('kc,inef,kncf,ae->ia', ampl.ph, t2, t2, op.vv)  # h.c.
            - 0.5 * einsum('kc,mnaf,kncf,mi->ia', ampl.ph, t2, t2, op.oo)  # h.c.
            # (2,7)
            - 1.0 * einsum('kc,kncf,imaf,mn->ia', ampl.ph, t2, t2, op.oo)
            + 1.0 * einsum('kc,knce,inaf,ef->ia', ampl.ph, t2, t2, op.vv)
        ))
    return IsrBlock(apply)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import sys
import warnings
import numpy as np
import scipy.linalg as la
import scipy.sparse.linalg as sla

from adcc import evaluate, lincomb
from adcc.AdcMatrix import AdcMatrixlike
from adcc.AmplitudeVector import AmplitudeVector

from .common import select_eigenpairs
from .preconditioner import JacobiPreconditioner
from .SolverStateBase import EigenSolverStateBase
from .explicit_symmetrisation import IndexSymmetrisation


class DavidsonState(EigenSolverStateBase):
    def __init__(self, matrix, guesses):
        super().__init__(matrix)
        self.residuals = None                   # Current residuals
        self.subspace_vectors = guesses.copy()  # Current subspace vectors
        self.algorithm = "davidson"


def default_print(state, identifier, file=sys.stdout):
    """
    A default print function for the davidson callback
    """
    from adcc.timings import strtime, strtime_short

    # TODO Use colour!

    if identifier == "start" and state.n_iter == 0:
        print("Niter n_ss  max_residual  time  Ritz values",
              file=file)
    elif identifier == "next_iter":
        time_iter = state.timer.current("iteration")
        fmt = "{n_iter:3d}  {ss_size:4d}  {residual:12.5g}  {tstr:5s}"
        print(fmt.format(n_iter=state.n_iter, tstr=strtime_short(time_iter),
                         ss_size=len(state.subspace_vectors),
                         residual=np.max(state.residual_norms)),
              "", state.eigenvalues[:7], file=file)
        if hasattr(state, "subspace_orthogonality"):
            print(33 * " " + "nonorth: {:5.3g}"
                  "".format(state.subspace_orthogonality))
    elif identifier == "is_converged":
        soltime = state.timer.total("iteration")
        print("=== Converged ===", file=file)
        print("    Number of matrix applies:   ", state.n_applies, file=file)
        print("    Total solver time:          ", strtime(soltime), file=file)
    elif identifier == "restart":
        print("=== Restart ===", file=file)


# TODO This function should be merged with eigsh
def davidson_iterations(matrix, state, max_subspace, max_iter, n_ep, n_block,
                        is_converged, which, callback=None, preconditioner=None,
                        preconditioning_method="Davidson", debug_checks=False,
                        residual_min_norm=None, explicit_symmetrisation=None,
                        max_subspace_iter=None):
    """Drive the davidson iterations

    Parameters
    ----------
    matrix
        Matrix to diagonalise
    state
        DavidsonState containing the eigenvector guess
    max_subspace : int
        Maximal subspace size
    max_iter : int
        Maximal number of iterations
    n_ep : int
        Number of eigenpairs to be computed
    n_block : int
        Davidson block size: the number of vectors that are added to the subspace
        in each iteration
    is_converged
        Function to test for convergence
    callback : callable, optional
        Callback to run after each iteration
    which : str, optional
        Which eigenvectors to converge to. Needs to be chosen such that
        it agrees with the selected preconditioner.
    preconditioner
        Preconditioner (type or instance)
    preconditioning_method : str, optional
        Precondititoning method. Valid values are "Davidson"
        or "Sleijpen-van-der-Vorst"
    debug_checks : bool, optional
        Enable some potentially costly debug checks
        (Loss of orthogonality etc.)
    residual_min_norm : float or NoneType, optional
        Minimal norm a residual needs to have in order to be accepted as
        a new subspace vector
        (defaults to 2 * len(matrix) * machine_expsilon)
    explicit_symmetrisation
        Explicit symmetrisation to apply to new subspace vectors before
        adding them to the subspace. Allows to correct for loss of index
        or spin symmetries (type or instance)
    max_subspace_iter : int, optional
        Maximum number of iterations for diagonalizing the subspace matrix
    """
    if preconditioning_method not in ["Davidson", "Sleijpen-van-der-Vorst"]:
        raise ValueError("Only 'Davidson' and 'Sleijpen-van-der-Vorst' "
                         "are valid preconditioner methods")
    if preconditioning_method == "Sleijpen-van-der-Vorst":
        raise NotImplementedError("Sleijpen-van-der-Vorst preconditioning "
                                  "not yet implemented.")

    if callback is None:
        def callback(state, identifier):
            pass

    # The problem size
    n_problem = matrix.shape[1]

    # The current subspace size == Number of guesses
    n_ss_vec = len(state.subspace_vectors)

    # Sanity checks for block size
    assert n_block >= n_ep and n_block <= n_ss_vec

    # The current subspace
    SS = state.subspace_vectors

    # The matrix A projected into the subspace
    # as a continuous array. Only the view
    # Ass[:n_ss_vec, :n_ss_vec] contains valid data.
    Ass_cont = np.empty((max_subspace, max_subspace))

    eps = np.finfo(float).eps
    if residual_min_norm is None:
        residual_min_norm = 2 * n_problem * eps

    callback(state, "start")
    state.timer.restart("iteration")

    with state.timer.record("projection"):
        # Initial application of A to the subspace
        Ax = evaluate(matrix @ SS)
        state.n_applies += n_ss_vec

    # Get the worksize view for the first iteration
    Ass = Ass_cont[:n_ss_vec, :n_ss_vec]

    # Initial projection of Ax onto the subspace exploiting the hermiticity
    with state.timer.record("projection"):
        for i in range(n_ss_vec):
            for j in range(i, n_ss_vec):
                Ass[i, j] = SS[i] @ Ax[j]
                if i != j:
                    Ass[j, i] = Ass[i, j]

    while state.n_iter < max_iter:
        state.n_iter += 1

        assert len(SS) >= n_block
        assert len(SS) <= max_subspace

        # Compute the which(== largest, smallest, ...) eigenpair of Ass
        # and the associated ritz vector as well as residual
        with state.timer.record("rayleigh_ritz"):
            if Ass.shape == (n_block, n_block):
                rvals, rvecs = la.eigh(Ass)  # Do a full diagonalisation
            else:
                # TODO Maybe play with precision a little here
                # TODO Maybe use previous vectors somehow
                v0 = None
                rvals, rvecs = sla.eigsh(Ass, k=n_block, which=which, v0=v0,
                                         maxiter=max_subspace_iter)

        with state.timer.record("residuals"):
            # Form residuals, A * SS * v - λ * SS * v = Ax * v + SS * (-λ*v)
            def form_residual(rval, rvec):
                coefficients = np.hstack((rvec, -rval * rvec))
                return lincomb(coefficients, Ax + SS, evaluate=True)
            residuals = [form_residual(rvals[i], v)
                         for i, v in enumerate(np.transpose(rvecs))]
            assert len(residuals) == n_block

            # Update the state's eigenpairs and residuals
            epair_mask = select_eigenpairs(rvals, n_ep, which)
            state.eigenvalues = rvals[epair_mask]
            state.residuals = [residuals[i] for i in epair_mask]
            state.residual_norms = np.array([np.sqrt(r @ r)
                                             for r in state.residuals])

        callback(state, "next_iter")
        state.timer.restart("iteration")
        if is_converged(state):
            # Build the eigenvectors we desire from the subspace vectors:
            state.eigenvectors = [lincomb(v, SS, evaluate=True)
                                  for i, v in enumerate(np.transpose(rvecs))
                                  if i in epair_mask]

            state.converged = True
            callback(state, "is_converged")
            state.timer.stop("iteration")
            return state

        if state.n_iter == max_iter:
            warnings.warn(la.LinAlgWarning(
                f"Maximum number of iterations (== {max_iter}) "
                "reached in davidson procedure."))
            state.eigenvectors = [lincomb(v, SS, evaluate=True)
                                  for i, v in enumerate(np.transpose(rvecs))
                                  if i in epair_mask]
            state.timer.stop("iteration")
            state.converged = False
            return state

        if n_ss_vec + n_block > max_subspace:
            callback(state, "restart")
            with state.timer.record("projection"):
                # The addition of the preconditioned vectors goes beyond max.
                # subspace size => Collapse first, ie keep current Ritz vectors
                # as new subspace
                SS = [lincomb(v, SS, evaluate=True) for v in np.transpose(rvecs)]
                state.subspace_vectors = SS
                Ax = [lincomb(v, Ax, evaluate=True) for v in np.transpose(rvecs)]
                n_ss_vec = len(SS)

                # Update projection of ADC matrix A onto subspace
                Ass = Ass_cont[:n_ss_vec, :n_ss_vec]
                for i in range(n_ss_vec):
                    for j in range(i, n_ss_vec):
                        Ass[i, j] = SS[i] @ Ax[j]
                        if i != j:
                            Ass[j, i] = Ass[i, j]
            # continue to add residuals to space

        with state.timer.record("preconditioner"):
            if preconditioner:
                if hasattr(preconditioner, "update_shifts"):
                    # Epsilon factor to make sure that 1 / (shift - diagonal)
                    # does not become ill-conditioned as soon as the shift
                    # approaches the actual diagonal values (which are the
                    # eigenvalues for the ADC(2) doubles part if the coupling
                    # block are absent)
                    rvals_eps = 1e-6
                    preconditioner.update_shifts(rvals - rvals_eps)

                preconds = evaluate(preconditioner @ residuals)
            else:
                preconds = residuals

            # Explicitly symmetrise the new vectors if requested
            if explicit_symmetrisation:
                explicit_symmetrisation.symmetrise(preconds)

        # Project the components of the preconditioned vectors away
        # which are already contained in the subspace.
        # Then add those, which have a significant norm to the subspace.
        with state.timer.record("orthogonalisation"):
            n_ss_added = 0
            for i in range(n_block):
                pvec = preconds[i]
                # Project out the components of the current subspace using
                # conventional Gram-Schmidt (CGS) procedure.
                # That is form (1 - SS * SS^T) * pvec = pvec + SS * (-SS^T * pvec)
                coefficients = np.hstack(([1], -(pvec @ SS)))
                pvec = lincomb(coefficients, [pvec] + SS, evaluate=True)
                pnorm = np.sqrt(pvec @ pvec)
                if pnorm < residual_min_norm:
                    continue
                # Perform reorthogonalisation if loss of orthogonality is
                # detected; this comes at the expense of computing n_ss_vec
                # additional scalar products but avoids linear dependence
                # within the subspace.
                with state.timer.record("reorthogonalisation"):
                    ss_overlap = np.array(pvec @ SS)
                    max_ortho_loss = np.max(np.abs(ss_overlap)) / pnorm
                    if max_ortho_loss > n_problem * eps:
                        # Update pvec by instance reorthogonalised against SS
                        # using a second CGS. Also update pnorm.
                        coefficients = np.hstack(([1], -ss_overlap))
                        pvec = lincomb(coefficients, [pvec] + SS, evaluate=True)
                        pnorm = np.sqrt(pvec @ pvec)
                        state.reortho_triggers.append(max_ortho_loss)
                if pnorm >= residual_min_norm:
                    # Extend the subspace
                    SS.append(evaluate(pvec / pnorm))
                    n_ss_added += 1
                    n_ss_vec = len(SS)

            if debug_checks:
                orth = np.array([[SS[i] @ SS[j] for i in range(n_ss_vec)]
                                 for j in range(n_ss_vec)])
                orth -= np.eye(n_ss_vec)
                state.subspace_orthogonality = np.max(np.abs(orth))
                if state.subspace_orthogonality > n_problem * eps:
                    warnings.warn(la.LinAlgWarning(
                        "Subspace in Davidson has lost orthogonality. "
                        "Max. deviation from orthogonality is {:.4E}. "
                        "Expect inaccurate results.".format(
                            state.subspace_orthogonality)
                    ))

        if n_ss_added == 0:
            state.timer.stop("iteration")
            state.converged = False
            state.eigenvectors = [lincomb(v, SS, evaluate=True)
                                  for i, v in enumerate(np.transpose(rvecs))
                                  if i in epair_mask]
            warnings.warn(la.LinAlgWarning(
                "Davidson procedure could not generate any further vectors for "
                "the subspace. Iteration cannot be continued like this and will "
                "be aborted without convergence. Try a different guess."))
            return state

        # Matrix applies for the new vectors
        with state.timer.record("projection"):
            Ax.extend(matrix @ SS[-n_ss_added:])
            state.n_applies += n_ss_added

        # Update the worksize view for the next iteration
        Ass = Ass_cont[:n_ss_vec, :n_ss_vec]

        # Project Ax onto the subspace, keeping in mind
        # that the values Ass[:-n_ss_added, :-n_ss_added] are already valid,
        # since they have been computed in the previous iterations already.
        with state.timer.record("projection"):
            for i in range(n_ss_vec - n_ss_added, n_ss_vec):
                for j in range(i + 1):
                    Ass[i, j] = SS[i] @ Ax[j]
                    if i != j:
                        Ass[j, i] = Ass[i, j]


def eigsh(matrix, guesses, n_ep=None, n_block=None, max_subspace=None,
          conv_tol=1e-9, which="SA", max_iter=70,
          callback=None, preconditioner=None,
          preconditioning_method="Davidson", debug_checks=False,
          residual_min_norm=None, explicit_symmetrisation=IndexSymmetrisation,
          max_subspace_iter=None):
    """Davidson eigensolver for ADC problems

    Parameters
    ----------
    matrix
        ADC matrix instance
    guesses : list
        Guess vectors (fixes also the Davidson block size)
    n_ep : int or NoneType, optional
        Number of eigenpairs to be computed
    n_block : int or NoneType, optional
        The solver block size: the number of vectors that are added to the subspace
        in each iteration
    max_subspace : int or NoneType, optional
        Maximal subspace size
    conv_tol : float, optional
        Convergence tolerance on the l2 norm of residuals to consider
        them converged
    which : str, optional
        Which eigenvectors to converge to (e.g. LM, LA, SM, SA)
    max_iter : int, optional
        Maximal number of iterations
    callback : callable, optional
        Callback to run after each iteration
    preconditioner
        Preconditioner (type or instance)
    preconditioning_method : str, optional
        Precondititoning method. Valid values are "Davidson"
        or "Sleijpen-van-der-Vorst"
    explicit_symmetrisation
        Explicit symmetrisation to apply to new subspace vectors before
        adding them to the subspace. Allows to correct for loss of index
        or spin symmetries (type or instance)
    debug_checks : bool, optional
        Enable some potentially costly debug checks
        (Loss of orthogonality etc.)
    residual_min_norm : float or NoneType, optional
        Minimal norm a residual needs to have in order to be accepted as
        a new subspace vector
        (defaults to 2 * len(matrix) * machine_expsilon)
    explicit_symmetrisation
        Explicit symmetrisation to apply to new subspace vectors before
        adding them to the subspace. Allows to correct for loss of index
        or spin symmetries (type or instance)
    max_subspace_iter : int, optional
        Maximum number of iterations for diagonalizing the subspace matrix
    """
    if not isinstance(matrix, AdcMatrixlike):
        raise TypeError("matrix is not of type AdcMatrixlike")
    for guess in guesses:
        if not isinstance(guess, AmplitudeVector):
            raise TypeError("One of the guesses is not of type AmplitudeVector")

    if preconditioner is not None and isinstance(preconditioner, type):
        preconditioner = preconditioner(matrix)

    if explicit_symmetrisation is not None and \
            isinstance(explicit_symmetrisation, type):
        explicit_symmetrisation = explicit_symmetrisation(matrix)

    if n_ep is None:
        n_ep = len(guesses)
    elif n_ep > len(guesses):
        raise ValueError(f"n_ep (= {n_ep}) cannot exceed the number of guess "
                         f"vectors (= {len(guesses)}).")

    if n_block is None:
        n_block = n_ep
    elif n_block < n_ep:
        raise ValueError(f"n_block (= {n_block}) cannot be smaller than the number "
                         f"of states requested (= {n_ep}).")
    elif n_block > len(guesses):
        raise ValueError(f"n_block (= {n_block}) cannot exceed the number of guess "
                         f"vectors (= {len(guesses)}).")

    if not max_subspace:
        # TODO Arnoldi uses this:
        # max_subspace = max(2 * n_ep + 1, 20)
        max_subspace = max(6 * n_ep, 20, 5 * len(guesses))
    elif max_subspace < 2 * n_block:
        raise ValueError(f"max_subspace (= {max_subspace}) needs to be at least "
                         f"twice as large as n_block (n_block = {n_block}).")
    elif max_subspace < len(guesses):
        raise ValueError(f"max_subspace (= {max_subspace}) cannot be smaller than "
                         f"the number of guess vectors (= {len(guesses)}).")

    def convergence_test(state):
        state.residuals_converged = state.residual_norms < conv_tol
        state.converged = np.all(state.residuals_converged)
        return state.converged

    if conv_tol < matrix.shape[1] * np.finfo(float).eps:
        warnings.warn(la.LinAlgWarning(
            "Convergence tolerance (== {:5.2g}) lower than "
            "estimated maximal numerical accuracy (== {:5.2g}). "
            "Convergence might be hard to achieve."
            "".format(conv_tol, matrix.shape[1] * np.finfo(float).eps)
        ))

    state = DavidsonState(matrix, guesses)
    davidson_iterations(matrix, state, max_subspace, max_iter,
                        n_ep=n_ep, n_block=n_block, is_converged=convergence_test,
                        callback=callback, which=which,
                        preconditioner=preconditioner,
                        preconditioning_method=preconditioning_method,
                        debug_checks=debug_checks,
                        residual_min_norm=residual_min_norm,
                        explicit_symmetrisation=explicit_symmetrisation,
                        max_subspace_iter=max_subspace_iter)
    return state


def jacobi_davidson(*args, **kwargs):
    return eigsh(*args, preconditioner=JacobiPreconditioner,
                 preconditioning_method="Davidson", **kwargs)


def davidson(*args, **kwargs):
    return eigsh(*args, preconditioner=None, **kwargs)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import copy

from pyscf import gto, scf
from matplotlib import pyplot as plt

mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvdz',
    unit="Bohr",
    verbose=0,
)

# Run normal SCF in pyscf
mf = scf.UHF(mol)
mf.conv_tol = 1e-13
mf.conv_tol_grad = 1e-8
mf.kernel()
print("Water SCF energy", mf.energy_tot())

# Make a core hole
mo0 = copy.deepcopy(mf.mo_coeff)
occ0 = copy.deepcopy(mf.mo_occ)
occ0[1][0] = 0.0  # beta core hole
dm = mf.make_rdm1(mo0, occ0)

mf_core = scf.UHF(mol)
mf_core.conv_tol = 1e-13
mf_core.conv_tol_grad = 1e-8
scf.addons.mom_occ(mf_core, mo0, occ0)
mf_core.kernel(dm)
del dm
print("Water core hole energy", mf_core.energy_tot())

# Run an adc2 calculation:
state = adcc.adc2(mf_core, n_states=4)

# Print results in a nice way
print()
print("  st  ex.ene. (au)         f     transition dipole moment (au)"
      "        state dip (au)")
for i, val in enumerate(state.excitation_energy):
    fmt = "{0:2d}  {1:12.8g} {2:9.3g}   [{3:9.3g}, {4:9.3g}, {5:9.3g}]"
    fmt += "   [{6:9.3g}, {7:9.3g}, {8:9.3g}]"
    print(state.kind[0], fmt.format(i, val, state.oscillator_strength[i],
                                    *state.transition_dipole_moment[i],
                                    *state.state_dipole_moment[i]))

state.plot_spectrum()
plt.savefig("pyscf_ccpvdz_xes_spectrum.pdf")
plt.show()

# Print timings summary:
print()
print(state.timer.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import os
import adcc
import matplotlib
import numpy as np

from pyscf import gto, scf

from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm

import h5py


def dump_matrix(basis, method):
    key = basis.replace("*", "s").replace("-", "").lower()
    fn = "water_" + key + "_" + method + ".hdf5"
    if os.path.isfile(fn):
        with h5py.File(fn, "r") as fp:
            return np.asarray(fp["adc_matrix"])

    mol = gto.M(
        atom="""O 0 0 0
                H 0 0 1.795239827225189
                H 1.693194615993441 0 -0.599043184453037""",
        basis=basis,
        unit="Bohr"
    )
    scfres = scf.RHF(mol)
    scfres.conv_tol = 1e-13
    scfres.kernel()

    mat = adcc.AdcMatrix(method, adcc.ReferenceState(scfres))
    key = basis.replace("*", "s").replace("-", "").lower()
    mat = mat.to_dense_matrix()

    fn = "water_" + key + "_" + method + ".hdf5"
    if not os.path.isfile(fn):
        with h5py.File(fn, "w") as fp:
            fp.create_dataset("adc_matrix", data=mat, compression="gzip")
    with h5py.File(fn, "r") as fp:
        return np.asarray(fp["adc_matrix"])


def plot_matrix(mtx, vrange=(-6, 2), sdline=None):
    plt.close()
    cmap = matplotlib.cm.YlOrRd
    cmap.set_bad("white", 1.)
    cmap.set_over(cmap(cmap.N), 1.)
    cmap.set_under("white", 1.)
    norm = LogNorm(vmin=10**min(vrange), vmax=10**max(vrange))

    ticks = []
    for i in range(min(vrange), max(vrange)):
        ticks += list(np.linspace(10**i, 10 * 10**i, 5, endpoint=False))
    ticks += [10**max(vrange)]

    img = plt.matshow(np.abs(mtx), cmap=cmap, norm=norm)
    plt.colorbar(img, ticks=ticks)
    plt.draw()

    if sdline:
        plt.axhline(y=40, xmin=0, xmax=mtx.shape[1],
                    color="grey", linewidth=0.7)
        plt.axvline(x=40, ymin=0, ymax=mtx.shape[1],
                    color="grey", linewidth=0.7)


def make_save_plot(basis, method, **kwargs):
    mtx = dump_matrix(basis, method)
    plot_matrix(mtx, **kwargs)

    key = basis.replace("*", "s").replace("-", "").lower()
    plt.savefig("matrix_water_" + method + "_" + key + ".png",
                bbox_inches="tight", dpi=200)


if __name__ == "__main__":
    # make_save_plot("sto-3g", "adc1", sdline=40)
    make_save_plot("sto-3g", "adc2", sdline=40)
    make_save_plot("sto-3g", "adc3", sdline=40)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='sto-3g',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-14
scfres.conv_tol_grad = 1e-10
scfres.kernel()

# Run an adc2 calculation:
singlets = adcc.adc2(scfres, n_singlets=5)
triplets = adcc.adc2(singlets.matrix, n_triplets=3)

print(singlets.describe())
print()
print(triplets.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2022 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np
import pytest
from numpy.testing import assert_allclose

import adcc

methods = ["adc0", "adc1", "adc2", "adc2x", "adc3"]
methods.extend([f"cvs-{m}" for m in methods])

_singlets = {
    "adc0": [0.99048446, 1.04450741, 1.15356339],
    "adc1": [0.48343609, 0.57420044, 0.60213699],
    "adc2": [0.47051314, 0.57255495, 0.59367339],
    "adc2x": [0.44511182, 0.55139508, 0.57146893],
    "adc3": [0.45197068, 0.55572706, 0.57861241],
    "cvs-adc0": [20.83623681, 20.99931574],
    "cvs-adc1": [20.10577515, 20.16512502],
    "cvs-adc2": [20.0045422,  20.08771799],
    "cvs-adc2x": [19.90528006, 19.9990468],
    "cvs-adc3": [19.93138526, 20.02055267],
}

_triplets = {
    "adc0": [0.99048446, 1.04450741, 1.15356339],
    "adc1": [0.40544164, 0.48351268, 0.52489129],
    "adc2": [0.40288477, 0.4913253,  0.52854722],
    "adc2x": [0.38557787, 0.48177562, 0.51336301],
    "adc3": [0.3924605,  0.4877098,  0.51755258],
    "cvs-adc0": [20.83623681, 20.99931574],
    "cvs-adc1": [20.04302084, 20.12099631],
    "cvs-adc2": [19.95732865, 20.05239094],
    "cvs-adc2x": [19.86572075, 19.97241019],
    "cvs-adc3": [19.89172578, 19.99386631],
}


def _residual_norms(state):
    residuals = [
        state.matrix @ v - e * v
        for e, v in zip(state.excitation_energy, state.excitation_vector)
    ]
    return np.array([r @ r for r in residuals])


@pytest.mark.skipif("pyscf" not in adcc.backends.available(),
                    reason="PySCF not found.")
@pytest.mark.parametrize("method", methods)
class TestSmoke:
    conv_tol = 1e-7

    def _run_scf_h2o_sto3g(self):
        # NOTE: cached_backend_hf not used because adcc.testdata
        # is not included in installation
        scfres = adcc.backends.run_hf(
            "pyscf", xyz="""
            O 0 0 0
            H 0 0 1.795239827225189
            H 1.693194615993441 0 -0.599043184453037""",
            basis="sto-3g", conv_tol=self.conv_tol / 100,
            conv_tol_grad=self.conv_tol / 10
        )
        return scfres

    def test_h2o_sto3g(self, method):
        scfres = self._run_scf_h2o_sto3g()
        if "cvs" in method:
            refstate = adcc.ReferenceState(scfres, core_orbitals=1)
            n_states = 2
        else:
            refstate = adcc.ReferenceState(scfres)
            n_states = 3
        state_singlets = adcc.run_adc(refstate, method=method,
                                      n_singlets=n_states, conv_tol=self.conv_tol)
        assert np.all(_residual_norms(state_singlets) < self.conv_tol)
        state_triplets = adcc.run_adc(refstate, method=method,
                                      n_triplets=n_states, conv_tol=self.conv_tol)
        assert np.all(_residual_norms(state_triplets) < self.conv_tol)
        assert_allclose(state_singlets.excitation_energy,
                        _singlets[method], atol=self.conv_tol * 10)
        assert_allclose(state_triplets.excitation_energy,
                        _triplets[method], atol=self.conv_tol * 10)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab

import adcc
from pyscf import gto, scf
from pyscf.solvent import PE

from scipy import constants
eV = constants.value("Hartree energy in eV")  # Hartree to eV

mol = gto.M(
    atom="""
    C          8.64800        1.07500       -1.71100
    C          9.48200        0.43000       -0.80800
    C          9.39600        0.75000        0.53800
    C          8.48200        1.71200        0.99500
    C          7.65300        2.34500        0.05500
    C          7.73200        2.03100       -1.29200
    H         10.18300       -0.30900       -1.16400
    H         10.04400        0.25200        1.24700
    H          6.94200        3.08900        0.38900
    H          7.09700        2.51500       -2.01800
    N          8.40100        2.02500        2.32500
    N          8.73400        0.74100       -3.12900
    O          7.98000        1.33100       -3.90100
    O          9.55600       -0.11000       -3.46600
    H          7.74900        2.71100        2.65200
    H          8.99100        1.57500        2.99500
    """,
    basis='sto-3g',
)

scfres = PE(scf.RHF(mol), {"potfile": "pna_6w.pot"})
scfres.conv_tol = 1e-8
scfres.conv_tol_grad = 1e-6
scfres.max_cycle = 250
scfres.kernel()

# model the solvent through perturbative corrections
state_pt = adcc.adc2(scfres, n_singlets=5, conv_tol=1e-5,
                     environment=['ptss', 'ptlr'])

# now model the solvent through linear-response coupling
# in the ADC matrix, re-using the matrix from previous run.
# This will modify state_pt.matrix
state_lr = adcc.run_adc(state_pt.matrix, n_singlets=5, conv_tol=1e-5,
                        environment='linear_response')

print(state_pt.describe())
print(state_lr.describe())
